#!/usr/bin/env python3

import sys
from itertools import chain
from pathlib import Path
from table import Table


HELP = """
$ base64bench_print.py old.txt old.new
"""


def main():
    script = sys.argv[0]
    args = sys.argv[1:]
    n = len(args)
    if "-h" in args or "--help" in args or (n != 2):
        print(HELP)
        return

    if n == 2:
        with open(args[0]) as f:
            old = parse(f.read())
        with open(args[1]) as f:
            new = parse(f.read())

        compare(old, new)


def compare(old, new):
    if old.metainfo != new.metainfo:
        print(old.metainfo)
        print(new.metainfo)
        return

    all = {impl: True for impl in chain(old.measurements.keys(), new.measurements.keys())}

    mi = old.metainfo
    title = '%s of %d bytes from %d input(s)' % (mi.mode, mi.volume, mi.inputs_count)

    table = Table()
    table.set_header(["implementation", "old [GB/s]", "new [GB/s]", "speedup"])
    for impl in all:
        try:
            old_speed = old.measurements[impl]
        except KeyError:
            old_speed = None

        try:
            new_speed = new.measurements[impl]
        except KeyError:
            new_speed = None

        if old_speed is not None and new_speed is not None:
            speedup = '%0.2fx' % (new_speed / old_speed)
            new_speed = '%0.2f' % new_speed
            old_speed = '%0.2f' % old_speed

        if old_speed is None:
            old_speed = '--'
            speedup   = '--'

        if new_speed is None:
            new_speed = '--'
            speedup   = '--'

        table.add_row([impl, old_speed, new_speed, speedup])

    print(table)


class Data:
    def __init__(self, mi):
        self.metainfo = mi
        self.measurements = {}


def parse(string):
    data = Data(parse_metainfo(string))
    for line in string.splitlines():
        line = line.rstrip()
        if not line:
            continue

        if not line.startswith('#'):
            impl, speed = parse_result(line)
            data.measurements[impl] = speed

    return data


def parse_metainfo(string):
    mi = Metainfo()
    mi.system = clip(string, '# current system detected as ', '.')
    mi.volume = int(clip(string, '# volume: ', ' bytes'))
    mi.max_length = int(clip(string, '# max length: ', ' bytes'))
    mi.inputs_count = int(clip(string, '# number of inputs: ', '\n'))

    modes = [
        ('# decode', 'decoding'),
    ]

    for frag, mode in modes:
        if frag in string:
            mi.mode = mode
            break

    return mi


class Metainfo:
    def __init__(self):
        self.system = None
        self.volume = None
        self.max_length = None
        self.inputs_count = None
        self.mode = None

    def __eq__(self, other):
        return vars(self) == vars(other)


def parse_result(line):
    name, sep, values = line.rpartition(':')
    name = name.strip()

    values = values.split()

    return (name, float(values[0]))


def clip(s, sep1, sep2):
    before, sep, after = s.partition(sep1)
    if sep == '':
        return None

    between, sep, after = after.partition(sep2)
    assert sep == sep2

    return between


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import re
import argparse

# Construct an argument parser
all_args = argparse.ArgumentParser()

# Add arguments to the parser
all_args.add_argument("-f", "--file", required=True,
   help="file name")
args = vars(all_args.parse_args())

filename = args['file']
with open(filename) as f:
    content = f.readlines()
table = []
currentrow = {}

datasets = set()
codecs = set()

for line in content:

    if line.startswith("convert"):
        m = re.search(r"\+(\w+)",line)
        if m is not None:
            codec = m.group(1)
            currentrow["codec"] = codec
            codecs.add(codec)
        m = re.search(r"/(\w+)[\.-]",line)
        if m is not None:
            rfile = m.group(1)
            currentrow["dataset"] = rfile
            datasets.add(rfile)
    m = re.search(r"\s([\.0-9]+) Gc/s",line)
    if m:
        v = float(m.group(1))
        currentrow["result"] = '{:#.2g}'.format(v)
        table.append(currentrow)
        currentrow = {}

favorite_kernels = ["icu", "llvm", "hoehrmann",  "cppcon2018", "u8u16", "utf8sse4", "utf8lut", "haswell", "arm64"]
kernels = []
s = "  "
for k in favorite_kernels:
    if k in codecs:
        kernels.append(k)
        s += " & " + k
s += " \\\\"
print(s)
def get(d, k):
    for x in table:
        if(x['codec'] == k) and (x['dataset'] == d):
            return x["result"]
datasetsorted=sorted(datasets)
for dataset in datasetsorted:
    s = dataset
    for k in kernels:
      s +=  " & " + get(dataset, k)
    s += " \\\\"
    print(s)



#!/usr/bin/env python3
#
# runs a test to ensure amalagamation works

import subprocess
import tempfile
import argparse
import datetime
import itertools
import math
import random


def test_one(args, features):
    print(f"testing with these features enabled: {', '.join(features)}")
    builddir = tempfile.TemporaryDirectory(delete=not args.keep_tmpdirs)
    t0 = datetime.datetime.now()
    subprocess.run(
        [
            "python3",
            "singleheader/amalgamate.py",
            "--source-dir",
            "src",
            "--include-dir",
            "include",
            "--output-dir",
            builddir.name,
            "--no-zip",
            "--no-readme",
        ]
        + [f"--with-{feature}" for feature in features],
        stdout=subprocess.DEVNULL,
        check=True,
    )
    print(f"amalgamated in {builddir.name}")
    if args.warnings:
        warnings = [
            "-Wall",
            "-Wextra",
            "-Werror",
            "-Wno-error=unused-function",
            "-Wno-error=unused-const-variable",
        ]
    else:
        warnings = []
    subprocess.run(
        [args.compiler, "-c", "simdutf.cpp", "-I", ".", f"-std=c++{args.standard}"]
        + warnings,
        check=True,
        cwd=builddir.name,
    )
    print(f"compiled in {builddir.name}")


allfeatures = ["utf8", "utf16", "utf32", "base64", "detect-enc", "ascii", "latin1"]


def test_combinations(args, include_count):
    total = math.comb(len(allfeatures), include_count)
    print(f"testing all {total} possible selections of {include_count} features")
    for i, combo in enumerate(itertools.combinations(allfeatures, include_count)):
        test_one(args, combo)
        print(f"tested {i + 1} of {total}")


def generate_subset():
    s = allfeatures
    for subset in itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    ):
        yield subset


def test_random(args, N):
    subsets = [s for s in generate_subset()]
    random.shuffle(subsets)
    if N > 0 and N < len(subsets):
        subsets = subsets[0:N]
    total = len(subsets)
    for i, subset in enumerate(subsets):
        print(f"testing {i + 1} of {total} subsets: {subset}")
        test_one(args, subset)


def main():
    parser = argparse.ArgumentParser(
        prog="amalgamation test", description="ensures that amalgamation works"
    )
    parser.add_argument("--standard", type=int, default=23)
    parser.add_argument("--keep-tmpdirs", action="store_true")
    parser.add_argument("--compiler", type=str, default="c++")
    parser.add_argument("--warnings", action="store_true")
    parser.add_argument(
        "--test-implicit-all",
        action="store_true",
        help="test giving no features, meaning all features are implicitly on",
    )
    parser.add_argument(
        "--test-explicit-all", action="store_true", help="test giving all features"
    )
    parser.add_argument(
        "--test-singles", action="store_true", help="test one feature at a time"
    )
    parser.add_argument(
        "--test-random",
        type=int,
        default=-1,
        help="tests up to the given amount of random subsets. 0 means all. negative means don't do this.",
    )
    parser.add_argument(
        "--test-pairs", action="store_true", help="tests all pairs of features"
    )
    parser.add_argument(
        "--test-triplets", action="store_true", help="tests all triplets of features"
    )
    parser.add_argument(
        "--test-all-but-one",
        action="store_true",
        help="tests all possible feature sets with one disabled",
    )

    args = parser.parse_args()

    if args.test_implicit_all:
        print("test all enabled, implicitly")
        test_one(args, [])

    if args.test_explicit_all:
        print("test all enabled, explicitly")
        test_one(args, allfeatures)

    if args.test_singles:
        print("test one single feature at a time")
        test_combinations(args, 1)

    if args.test_random >= 0:
        print("test random")
        test_random(args, args.test_random)

    if args.test_pairs:
        print("test all pairs")
        test_combinations(args, 2)

    if args.test_triplets:
        print("test all triplets")
        test_combinations(args, 3)

    if args.test_all_but_one:
        print("test all but one single feature at a time")
        test_combinations(args, len(allfeatures) - 1)

    print("all seems to have gone well. bye bye!")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import sys


def format_array(array):
    result = []
    for value in array:
        if value < 0 or value == 0x80:
            result.append('0x80')
        else:
            result.append(str(value))

    return ', '.join(result)


def assure_array_length(array, size, value = 0x80):
    while len(array) < size:
        array.append(value)


CPP_1_2 = """
  // 1 byte for length, 16 bytes for mask
  const uint8_t pack_1_2_utf8_bytes[256][17] = {
%(rows)s
  };
"""


# For all patterns the 0th element of shuffle is 0.
# We may reuse that entry to store length, but it would
# require some changes in C++ code.
def shuffle_for_conversion_1_or_2_utf8_bytes(file):
  rows = []
  indent = (' ' * 4)
  for shuffle, size in shuffle_for_conversion_1_or_2_utf8_bytes_aux():
    array_str = []
    for value in [size] + shuffle:
      if value == 0x80:
        array_str.append('0x80')
      else:
        array_str.append(str(value))

    array = ','.join(array_str)
    rows.append(f'{indent}{{{array}}}')

  file.write(CPP_1_2 % {'rows': ',\n'.join(rows)})


def shuffle_for_conversion_1_or_2_utf8_bytes_aux():
  # We process 8 x 16-bit word
  # a bit one indices a word having values 0x00..0x7f (produces a single UTF-8 byte)
  # a bit zero indices a word having values 0x0080..0x7ff (produces two UTF-8 bytes)

  # Our input is a 16-bit word in form hhggffeeddccbbaa -- the bits are doubled
  # (h - MSB, a - LSB). In a C++ code we transform it using the following formula:
  #
  # in = hhggffeeddccbbaa
  # t0 = in & 0x5555       // t0 = 0h0g0f0e0d0c0b0a
  # t1 = t0 >> 7           // t1 = 00000000h0g0f0e0
  # t2 = (t0 | t1) & 0xff  // t2 =         hdgcfbea

  for mask in range(256):
    def getbit(k):
      return (mask & (1 << k) != 0)

    a = getbit(0)
    b = getbit(2)
    c = getbit(4)
    d = getbit(6)
    e = getbit(1)
    f = getbit(3)
    g = getbit(5)
    h = getbit(7)

    shuffle = []
    for word_index, bit in enumerate([a, b, c, d, e, f, g, h]):
      if bit: # 1 byte
        shuffle.append(word_index * 2)
      else: # 2 bytes
        shuffle.append(word_index * 2 + 1)
        shuffle.append(word_index * 2)

    output_bytes = len(shuffle)
    while (len(shuffle) < 16):
      shuffle.append(0x80)

    yield (shuffle, output_bytes)


CPP_1_2_3 = """
  // 1 byte for length, 16 bytes for mask
  const uint8_t pack_1_2_3_utf8_bytes[256][17] = {
%(rows)s
  };
"""

def shuffle_for_conversion_1_2_3_utf8_bytes(file):
  rows = []
  indent = (' ' * 4)
  for shuffle, size in shuffle_for_conversion_1_2_3_utf8_bytes_aux():
    array_str = []
    for value in [size] + shuffle:
      if value == 0x80:
        array_str.append('0x80')
      else:
        array_str.append(str(value))

    array = ','.join(array_str)
    rows.append(f'{indent}{{{array}}}')

  file.write(CPP_1_2_3 % {'rows': ',\n'.join(rows)})


def shuffle_for_conversion_1_2_3_utf8_bytes_aux():
  # There are two 8-bit bitmask telling how many bytes each word produces (1, 2 or 3).
  # mask1 = ddccbbaa -- output exactly one byte (d - MSB, a - LSB)
  # mask2 = hhggffee -- output one or two bytes

  # Please note that each bit is duplicated. In final form these bits are interleaved:
  # mask  = (mask1 & 0x5555) | (mask2 & 0xaaaa)
  #       = hdgcfbea

  # Each two-bit subword decides how many bytes will be copied from a 32-bit word of register:
  # | e | a | ea |
  # +---+---+----+-------
  # | 0 | 0 |  0 |  3 bytes
  # | 0 | 1 |  1 |  -- such combination will never come from C++ code, it has no sense
  # | 1 | 0 |  2 |  2 bytes
  # | 1 | 1 |  3 |  1 byte

  for mask in range(256):
    empty = 0x80
    shuffle = []
    for i in range(4):
      subword = mask & 0b11
      mask >>= 2

      if subword == 0:
        shuffle.append(i*4 + 2)
        shuffle.append(i*4 + 3)
        shuffle.append(i*4 + 1)
      elif subword == 3:
        shuffle.append(i*4 + 0)
      elif subword == 2:
        shuffle.append(i*4 + 3)
        shuffle.append(i*4 + 1)

    output_bytes = len(shuffle)
    while (len(shuffle) < 16):
      shuffle.append(empty)

    yield (shuffle, output_bytes)


CPP_EXPAND_SURROGATES = """
  // 2x16 bytes for masks, dwords_consumed
  const uint8_t expand_surrogates[256][33] = {
%(rows)s
  };
"""


def shuffle_for_expanding_surrogate_pairs(file):
  rows = []
  indent = (' ' * 4)
  for shuffle, dwords_consumed in shuffle_for_expanding_surrogate_pairs_aux():

    # If we consume, say 6 dwords of 8, then anyway the C++ conversion
    # routing convert 2 extra dwords (zeroed) into 2 UTF-8 bytes. Thus
    # we have to subtract this zero_dwords from saved bytes, to get
    # the real number of output bytes.
    zero_dwords = 8 - dwords_consumed;

    assert len(shuffle) == 32
    rows.append('%s{%s}' % (indent, format_array(shuffle + [zero_dwords])))

  file.write(CPP_EXPAND_SURROGATES % {'rows': ',\n'.join(rows)})


# Our input 8-bit bitmask informs which word contains a surrogate (low or high one).
# At this point we do not need to know which is which, as we assume that word
# expansion is done after validation. (Let's assume L - low surrogate, H - high
# surrogate, V - any valid non-surrogate word).
#
# Example 1: bitmask 1001'1110 describes a sequence V-L-H-L-H-V-V-? -- the last
# surrogate word might be either L or H, we'll ignore it. Two adjacent bits
# are expected to contain low & high surrogates
#
# Example 2: bitmask 0011'0110 describes a sequence V-L-K-V-L-H-V-V.
#
# Example 3: bitmask 0000'0001 is not valid --- sole surrogate word must not start
# a chunk of string, and C++ takes care not to pass such wrong input.
#
# Example 4: bitmask 0000'1110 is not valid too
#
# We expand all words into  32-bit lanes, spanning two SSE registers.
def shuffle_for_expanding_surrogate_pairs_aux():

  def shuffle_mask(mask):
    result = []
    prev = 'V'
    dwords_consumed = 0
    for i in range(8):
      bit = bool(mask & (1 << i))
      if bit:
        if prev == 'V':
          curr = 'L'
        elif prev == 'L':
          curr = 'H'
        elif prev == 'H':
          curr = 'L'

        result.append(2*i + 0)
        result.append(2*i + 1)

        if curr == 'L':
          dwords_consumed += 1

      else:
        if prev == 'V':
          curr = 'V'
        elif prev == 'L':
          raise ValueError('invalid sequence')
        elif prev == 'H':
          curr = 'V'

        result.append(2*i + 0)
        result.append(2*i + 1)
        result.append(-1)
        result.append(-1)

        dwords_consumed += 1

      prev = curr
    #for

    if curr == 'L': # a sole low surrogate word at the end, discard it (C++ code deals with this case)
        del result[-1]
        del result[-1]
        dwords_consumed -= 1

    while len(result) < 32:
        result.append(-1)

    return result, dwords_consumed

  invalid = 0

  # our input is in form: hdgcfbea
  # we need bits in seq:  hgfedcba
  def as_mask(x):
    def bit(k):
      return int(bool((1 << k) & x))

    return bit(0) \
         | (bit(2) << 1) \
         | (bit(4) << 2) \
         | (bit(6) << 3) \
         | (bit(1) << 4) \
         | (bit(3) << 5) \
         | (bit(5) << 6) \
         | (bit(7) << 7)


  if False:
      print('{:08b}'.format(as_mask(0x85)))
      shuffle_mask(as_mask(0x85))
      sys.exit(1)

  for x in range(256):
    mask = as_mask(x)

    try:
      yield shuffle_mask(mask)
    except ValueError:
      yield (([-1] * 32), 0)


CPP_UCS4_TO_UTF8 = """
  struct UCS4_to_UTF8 {
    uint8_t shuffle[16];
    uint8_t const_bits_mask[16];
    uint8_t output_bytes;
  };

  static_assert(sizeof(UCS4_to_UTF8) == 33, "Structure must be packed");

  const UCS4_to_UTF8 ucs4_to_utf8[256] = {
%(rows)s
  };
"""

"""
The input is 8-bit mask: geca'hfdb. Two-bit words: ab, cd, ef, gh
encodes how many UTF-8 bytes are store in each dword of an SSE
register:
- 00 - 1 byte
- 01 - 2 bytes
- 10 - 3 bytes
- 11 - 4 bytes

We output 3 values:
- a shuffle mask to extract UTF-8 bytes,
- mask to complete UTF-8 format,
- the total number of UTF-8 bytes.
"""
def ucs4_to_utf8(file):
    rows = []
    indent = (' ' * 4)
    for shuffle, const_bits_mask, output_bytes in ucs4_to_utf8_aux():
        #print(output_bytes)
        rows.append('%s{{%s}, {%s}, %d}' % (indent,
                                            format_array(shuffle),
                                            format_array(const_bits_mask),
                                            output_bytes))

    file.write(CPP_UCS4_TO_UTF8 % {'rows': ',\n'.join(rows)})


def ucs4_to_utf8_aux():
    for x in range(256):
        shuffle     = []
        utf8bits    = []
        output_bytes = 0

        def bit(k):
            return int(bool((1 << k) & x))

        def code(bit1, bit0):
            return 2*bit1 + bit0

        ab = code(bit(1), bit(0))
        cd = code(bit(5), bit(4))
        ef = code(bit(3), bit(2))
        gh = code(bit(7), bit(6))

        for i, count in enumerate([ab, cd, ef, gh]):
            if count == 0:
                shuffle.append(4*i + 0)

                utf8bits.append(0x00)
                utf8bits.append(0x00)
                utf8bits.append(0x00)
                utf8bits.append(0x00)

                output_bytes += 1
            elif count == 1:
                shuffle.append(4*i + 1)
                shuffle.append(4*i + 0)

                utf8bits.append(0b10000000)
                utf8bits.append(0b11000000)
                utf8bits.append(0x00)
                utf8bits.append(0x00)

                output_bytes += 2
            elif count == 2:
                shuffle.append(4*i + 2)
                shuffle.append(4*i + 1)
                shuffle.append(4*i + 0)

                utf8bits.append(0b10000000)
                utf8bits.append(0b10000000)
                utf8bits.append(0b11100000)
                utf8bits.append(0x00)

                output_bytes += 3
            elif count == 3:
                shuffle.append(4*i + 3)
                shuffle.append(4*i + 2)
                shuffle.append(4*i + 1)
                shuffle.append(4*i + 0)

                utf8bits.append(0b10000000)
                utf8bits.append(0b10000000)
                utf8bits.append(0b10000000)
                utf8bits.append(0b11110000)

                output_bytes += 4
            else:
                assert False

        assure_array_length(shuffle, 16, 0x80)
        assert len(utf8bits) == 16
        assert len(shuffle) == 16

        yield (shuffle, utf8bits, output_bytes)



CPP_HEADER = """// file generated by scripts/sse_convert_utf16_to_utf8.py
#ifndef SIMDUTF_UTF16_TO_UTF8_TABLES_H
#define SIMDUTF_UTF16_TO_UTF8_TABLES_H

namespace simdutf {
namespace {
namespace tables {
namespace utf16_to_utf8 {
"""

CPP_FOOTER = """} // utf16_to_utf8 namespace
} // tables namespace
} // unnamed namespace
} // namespace simdutf

#endif // SIMDUTF_UTF16_TO_UTF8_TABLES_H
"""

def main():
  with open('utf16_to_utf8_tables.h', 'wt') as f:
    f.write(CPP_HEADER)
    shuffle_for_conversion_1_or_2_utf8_bytes(f)
    shuffle_for_conversion_1_2_3_utf8_bytes(f)
    shuffle_for_expanding_surrogate_pairs(f)
    ucs4_to_utf8(f)
    f.write(CPP_FOOTER)


if __name__ == '__main__':
    main()




#!/usr/bin/env python3
import subprocess
import re
import os
import argparse
import sys
from typing import List, Tuple

def get_git_files(limit_to_src_include: bool) -> List[str]:
    """Retrieve list of .cpp and .h files checked into the Git repository."""
    try:
        patterns = ['src/*.cpp', 'src/*.h', 'include/*.cpp', 'include/*.h'] if limit_to_src_include else ['*.cpp', '*.h']
        result = subprocess.run(
            ['git', 'ls-files'] + patterns,
            capture_output=True,
            text=True,
            check=True
        )
        files = result.stdout.splitlines()
        if limit_to_src_include:
            files = [f for f in files if f.startswith(('src/', 'include/'))]
        return files
    except subprocess.CalledProcessError as e:
        print(f"Error running git ls-files: {e}")
        return []

def is_preprocessor_directive(line: str) -> bool:
    """Check if the line contains a preprocessor directive (not in a comment)."""
    line = re.sub(r'//.*$', '', line)
    return bool(re.match(r'^\s*#', line))

def extract_directive(lines: List[str], line_number: int) -> Tuple[str, str]:
    """Extract directive type and condition (if any) from a preprocessor line."""
    line = lines[line_number].strip()
    # Sometimes, the condition may span multiple lines due to backslashes
    # Handle multi-line conditions for #if directives
    # We recursively gather lines if they end with a backslash
    def build_full_condition(comment: str, line_number: int) -> str:
        if comment.endswith('\\') and line_number + 1 < len(lines):
            line_number += 1
            return comment[:-1].strip() + ' ' + build_full_condition(lines[line_number].strip(), line_number)
        else:
            return comment.strip()
    if line.startswith('#if '):
        return 'if', build_full_condition(line[4:].strip(), line_number)
    elif line.startswith('#ifdef '):
        return 'ifdef', build_full_condition(line[7:].strip(), line_number)
    elif line.startswith('#ifndef '):
        return 'ifndef', build_full_condition(line[8:].strip(), line_number)
    elif line.startswith('#endif'):
        match = re.match(r'#endif\s*(?://\s*(.*))?$', line)
        ## In some instances, the comment may be on the next line(s)
        ## We use the following loop to gather such comments
        ## The heuristic is that the next line starts with '//' and contains 'SIMDUTF_FEATURE'
        ## Note that if we don't check for SIMDUTF_FEATURE, we may pick up unrelated comments.
        comment = match.group(1).strip() if match and match.group(1) else ''
        while line_number + 1 < len(lines):
            next_line = lines[line_number + 1].strip()
            if next_line.startswith('//') and 'SIMDUTF_FEATURE' in next_line:
                comment_part = next_line[2:].strip()
                comment = (comment + ' ' + comment_part).strip() if comment else comment_part
                line_number += 1
            else:
                break
        return 'endif', comment
    return '', ''

def is_feature_if_directive(condition: str) -> bool:
    """Check if the #if condition contains a macro starting with SIMDUTF_FEATURE."""
    return bool(re.search(r'\bSIMDUTF_FEATURE\w*\b', condition))

def check_file(filename: str) -> List[str]:
    """Check #if directives with SIMDUTF_FEATURE macros for matching #endif comments."""
    errors = []
    stack = []  # Stack of (directive, condition, is_feature, line_number)
    statistics = {'if': 0, 'ifdef': 0, 'ifndef': 0, 'endif': 0}

    try:
        with open(filename, 'r', encoding='utf-8') as f:
            lines = f.readlines()

        for line_number, line in enumerate(lines):
            if not is_preprocessor_directive(line):
                continue

            directive, condition = extract_directive(lines, line_number)
            if not directive:
                continue
            statistics[directive] += 1

            if directive in ('if', 'ifdef', 'ifndef'):
                is_feature = directive == 'if' and is_feature_if_directive(condition)
                stack.append((directive, condition, is_feature, line_number))
            elif directive == 'endif' and stack:
                orig_directive, orig_condition, is_feature, orig_line = stack.pop()
                if is_feature:
                    if not condition:
                        errors.append(
                            f"\033[31mERROR: {filename}:{line_number}: #endif missing comment (expected '{orig_condition}')\033[0m"
                        )
                    elif not re.match(r'SIMDUTF_FEATURE\w*', condition):
                        errors.append(
                            f"\033[31mERROR: {filename}:{line_number}: #endif comment '{condition}' does not start with SIMDUTF_FEATURE "
                            f"(expected '{orig_condition}')\033[0m"
                        )
                    elif condition != orig_condition:
                        errors.append(
                            f"\033[31mERROR: {filename}:{line_number}: #endif comment '{condition}' does not match "
                            f"#if condition '{orig_condition}' at line {orig_line}\033[0m"
                        )

        # Check for unclosed feature-related #if directives
        for directive, condition, is_feature, line in stack:
            if is_feature:
                red_warning = f"\033[31mWARNING: {filename}:{line}: Unclosed #if '{condition}' at end of file\033[0m"
                errors.append(red_warning)

        return errors, statistics

    except FileNotFoundError:
        return [f"{filename}: File not found"], statistics
    except UnicodeDecodeError:
        return [f"{filename}: Unable to decode file (invalid encoding)"], statistics

def main():
    """Main function to check .cpp and .h files for SIMDUTF feature #if...#endif pairs."""
    parser = argparse.ArgumentParser(description="Check #if...#endif pairs for SIMDUTF feature macros in .cpp and .h files.")
    parser.add_argument('--limit', action='store_true', help="Limit checks to files in src/ and include/ directories")
    args = parser.parse_args()

    files = get_git_files(args.limit)
    if not files:
        print("No .cpp or .h files found in the Git repository" + (" in src/ or include/" if args.limit else "") + ".")
        sys.exit(1)  # Exit with error code if no files are found

    all_errors = []
    statistics = {'if': 0, 'ifdef': 0, 'ifndef': 0, 'endif': 0}
    for filename in files:
        errors, file_stats = check_file(filename)
        all_errors.extend(errors)
        for key in statistics:
            statistics[key] += file_stats.get(key, 0)

    if all_errors:
        print("Errors found:")
        for error in all_errors:
            print(error)
        sys.exit(1)  # Exit with error code if errors are found
    else:
        for key, count in statistics.items():
            print(f"Total #{key} directives: {count}")
        print("All #if...#endif pairs for SIMDUTF feature macros are correctly matched with valid comments.")
        sys.exit(0)  # Exit with success code

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
from common import *


def is_bit_set(mask, i):
    return (mask & ( 1<<i )) == ( 1<<i )

# computes the location of the 0 bits (index starts at zero)
def compute_locations(mask):
    answer = []
    i = 0
    while( (mask >> i) > 0 ):
        if(is_bit_set(mask,i)):
            answer.append(i)
        i += 1
    return answer


# computes the gaps between the 1, assuming we had an initial 1
def compute_code_point_size(mask):
    positions = compute_locations(mask)
    answer = []
    oldx = -1
    for i in range(len(positions)):
        x = positions[i]
        answer.append(x-oldx)
        oldx = x
    return answer

## check that we have 6 1-2 byte (at least)
def easy_case12(code_point_size):
    if(len(code_point_size)<6):
        return False
    return max(code_point_size[:6])<=2

## check that we have 4 1-2-3 byte (at least)
def easy_case123(code_point_size):
    if(len(code_point_size)<4):
        return False
    return max(code_point_size[:4])<=3

## check that we have 4 1-2-3 byte (at least)
def easy_case1234(code_point_size):
    if(len(code_point_size)<3):
        return False
    return max(code_point_size[:3])<=4

def grab_easy_case12_code_point_size(code_point_size):
    return code_point_size[:6]

def grab_easy_case123_code_point_size(code_point_size):
    return code_point_size[:4]

def grab_easy_case1234_code_point_size(code_point_size):
    return code_point_size[:3]

def buildshuf12_twobytes(sizes):
    answer = [0 for i in range(16)]
    pos = 0
    for i in range(len(sizes)):
        if(sizes[i] == 1):
            answer[2*i] = pos
            answer[2*i+1] = 0xff
            pos += 1
        else:
            answer[2*i] = pos + 1
            answer[2*i+1] = pos
            pos += 2
    return answer

def buildshuf123_threebytes(sizes):
    answer = [0 for i in range(16)]
    pos = 0
    for i in range(len(sizes)): # 4 * 4 = 16
        if(sizes[i] == 1):
            answer[4*i] = pos
            answer[4*i+1] = 0xff
            answer[4*i+2] = 0xff
            answer[4*i+3] = 0xff
            pos += 1
        elif(sizes[i] == 2):
            answer[4*i] = pos + 1
            answer[4*i+1] = pos
            answer[4*i+2] = 0xff
            answer[4*i+3] = 0xff
            pos += 2
        else: # must be three
            answer[4*i] = pos + 2
            answer[4*i+1] = pos + 1
            answer[4*i+2] = pos
            answer[4*i+3] = 0xff
            pos += 3
    return answer

def buildshuf1234_fourbytes(sizes):
    answer = [0 for i in range(16)]
    pos = 0
    for i in range(len(sizes)): # 3 * 4 = 12
        if(sizes[i] == 1):
            answer[4*i] = pos
            answer[4*i+1] = 0xff
            answer[4*i+2] = 0xff
            answer[4*i+3] = 0xff
            pos += 1
        elif(sizes[i] == 2):
            answer[4*i] = pos + 1
            answer[4*i+1] = pos
            answer[4*i+2] = 0xff
            answer[4*i+3] = 0xff
            pos += 2
        elif(sizes[i] == 3):
            answer[4*i] = pos + 2
            answer[4*i+1] = pos + 1
            answer[4*i+2] = pos
            answer[4*i+3] = 0xff
            pos += 3
        else: # must be four
            answer[4*i] = pos + 3
            answer[4*i+1] = pos + 2
            answer[4*i+2] = pos + 1
            answer[4*i+3] = pos
            pos += 4
    return answer


def main():
  easycase12 = set()
  easycase123 = set()
  easycase1234 = set()
  for x in range(1<<12):
    sizes = compute_code_point_size(x)
    if(easy_case12(sizes)):
        z1 = grab_easy_case12_code_point_size(sizes)
        easycase12.add(tuple(z1))
    elif(easy_case123(sizes)):
        z1 = grab_easy_case123_code_point_size(sizes)
        easycase123.add(tuple(z1))
    elif(easy_case1234(sizes)):
        z1 = grab_easy_case1234_code_point_size(sizes)
        easycase1234.add(tuple(z1))
  easycase12sorted = [x for x in easycase12]
  easycase12sorted.sort()
  easycase123sorted = [x for x in easycase123]
  easycase123sorted.sort()
  easycase1234sorted = [x for x in easycase1234]
  easycase1234sorted.sort()

  print("#include <cstdint>")
  allshuf = [buildshuf12_twobytes(z) for z in  easycase12sorted] + [buildshuf123_threebytes(z) for z in  easycase123sorted] + [buildshuf1234_fourbytes(z) for z in  easycase1234sorted]
  print("const uint8_t shufutf8["+str(len(easycase12sorted+easycase123sorted+easycase1234sorted))+"][16] = ")
  print(cpp_arrayarray_initializer(allshuf), end=";\n")
  print("/* number of two bytes : "+ str(len(easycase12sorted))+ " */")
  print("/* number of two + three bytes : "+ str(len(easycase12sorted+easycase123sorted))+ " */")
  print("/* number of two + three + four bytes : "+ str(len(easycase12sorted+easycase123sorted+easycase1234sorted))+ " */")
  c = 0
  index = {}
  for t in easycase12sorted + easycase123sorted + easycase1234sorted:
      index[t] = c
      c = c + 1
  arrg=[]
  for x in range(1<<12):
    sizes = compute_code_point_size(x)
    if(easy_case12(sizes)):
        z1 = grab_easy_case12_code_point_size(sizes)
        idx = index[tuple(z1)]
        s = sum(z1)
        arrg.append((idx,s))
    elif(easy_case123(sizes)):
        z1 = grab_easy_case123_code_point_size(sizes)
        idx = index[tuple(z1)]
        s = sum(z1)
        arrg.append((idx,s))
    elif(easy_case1234(sizes)):
        z1 = grab_easy_case1234_code_point_size(sizes)
        idx = index[tuple(z1)]
        s = sum(z1)
        arrg.append((idx,s))
    else:
        # we are in error, use a bogus index
        arrg.append((209,12))
  print("const uint8_t utf8bigindex["+str(len(arrg))+"][2] = ")
  print(cpp_arrayarray_initializer(arrg), end=";\n")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Creates the amalgamated source files.

import argparse
import sys
import os.path
import subprocess
import os
import pathlib
import re
import shutil
import datetime

if sys.version_info[0] < 3:
    sys.stdout.write("Sorry, requires Python 3.x or better\n")
    sys.exit(1)


SCRIPTPATH = os.path.dirname(os.path.abspath(sys.argv[0]))
PROJECTPATH = os.path.dirname(SCRIPTPATH)


class Context:
    pass


context = Context()


def main():
    (args, enabled_features) = parse_args()

    print(f"SCRIPTPATH={SCRIPTPATH} PROJECTPATH={PROJECTPATH}")
    print("We are about to amalgamate all simdutf files into one source file.")
    print("See https://www.sqlite.org/amalgamation.html and https://en.wikipedia.org/wiki/Single_Compilation_Unit for rationale.")

    context.args             = args
    context.enabled_features = enabled_features
    context.zipname          = 'singleheader.zip'
    context.timestamp        = get_timestamp()
    if enabled_features != known_features:
        context.read_file = filter_features
    else:
        context.read_file = read_file

    print(f"timestamp is {context.timestamp}")

    create_files()
    if args.zip:
        create_zip()

    print("Done with all files generation.")
    print(f"Files have been written to directory: {context.args.output_dir}")

    if args.readme:
        print_instructions()


known_features = {
    'SIMDUTF_FEATURE_DETECT_ENCODING',
    'SIMDUTF_FEATURE_LATIN1',
    'SIMDUTF_FEATURE_ASCII',
    'SIMDUTF_FEATURE_BASE64',
    'SIMDUTF_FEATURE_UTF8',
    'SIMDUTF_FEATURE_UTF16',
    'SIMDUTF_FEATURE_UTF32',
}


def parse_args():
    p = argparse.ArgumentParser("SIMDUTF tool for amalgmation")
    p.add_argument("--source-dir",
                   default=os.path.join(PROJECTPATH, "src"),
                   metavar="SRC",
                   help="Source dir")
    p.add_argument("--include-dir",
                   default=os.path.join(PROJECTPATH, "include"),
                   metavar="INC",
                   help="Include dir")
    p.add_argument("--output-dir",
                   default=os.path.join(SCRIPTPATH),
                   metavar="DIR",
                   help="Output directory")
    p.add_argument("--no-zip",
                   default=True,
                   action='store_false',
                   dest='zip',
                   help="Do not create .zip file")
    p.add_argument("--no-readme",
                   default=True,
                   action='store_false',
                   dest='readme',
                   help="Do not show readme after creating files")
    p.add_argument("--with-utf8",
                   default=None,
                   action='store_true',
                   help="Include UTF-8 support")
    p.add_argument("--with-utf16",
                   default=None,
                   action='store_true',
                   help="Include UTF-16 support")
    p.add_argument("--with-utf32",
                   default=None,
                   action='store_true',
                   help="Include UTF-32 support")
    p.add_argument("--with-base64",
                   default=None,
                   action='store_true',
                   help="Include Base64 support")
    p.add_argument("--with-detect-enc",
                   default=None,
                   action='store_true',
                   help="Include encoding detection support")
    p.add_argument("--with-ascii",
                   default=None,
                   action='store_true',
                   help="Include ASCII support")
    p.add_argument("--with-latin1",
                   default=None,
                   action='store_true',
                   help="Include Latin1 support")
    p.add_argument("--debug-sources",
                   default=False,
                   action='store_true',
                   help="Include exact source location in amalgamated sources")

    args = p.parse_args()

    items = (
        ("AMALGAMATE_SOURCE_PATH", "source_dir"),
        ("AMALGAMATE_INCLUDE_PATH", "include_dir"),
        ("AMALGAMATE_OUTPUT_PATH", "output_dir"),
    )
    for var, attribute in items:
        if var in os.environ:
            val = os.environ[var]
            print(f"using env variable {var}={val}")
            setattr(args, attribute, val)

    enabled_features = set()
    if args.with_utf8:
        enabled_features.add('SIMDUTF_FEATURE_UTF8')
    if args.with_utf16:
        enabled_features.add('SIMDUTF_FEATURE_UTF16')
    if args.with_utf32:
        enabled_features.add('SIMDUTF_FEATURE_UTF32')
    if args.with_base64:
        enabled_features.add('SIMDUTF_FEATURE_BASE64')
    if args.with_detect_enc:
        enabled_features.add('SIMDUTF_FEATURE_DETECT_ENCODING')
    if args.with_ascii:
        enabled_features.add('SIMDUTF_FEATURE_ASCII')
    if args.with_latin1:
        enabled_features.add('SIMDUTF_FEATURE_LATIN1')

    if not enabled_features:
        enabled_features = set(known_features)

    return (args, enabled_features)


def doinclude(file, line):
    for directory in [context.args.include_dir, context.args.source_dir]:
        path = os.path.join(directory, file)
        if os.path.exists(path):
            # generic includes are included multiple times
            if re.match('.*generic/.*.h', file):
                dofile(directory, file)
            # begin/end_implementation are also included multiple times
            elif re.match('.*/begin.h', file):
                dofile(directory, file)
            elif re.match('.*/end.h', file):
                dofile(directory, file)
            elif file not in context.found_includes:
                context.found_includes.add(file)
                dofile(directory, file)
            else:
                pass

            return

    # If we don't recognize it, just emit the #include
    print(line, file=context.fid)


def dofile(prepath, filename):
    file = os.path.join(prepath, filename)
    fid  = context.fid
    RELFILE = os.path.relpath(file, PROJECTPATH)
    # Last lines are always ignored. Files should end by an empty lines.
    print(f"/* begin file {RELFILE} */", file=fid)
    includepattern = re.compile(r'^\s*#\s*include (<|")(.*)(>|")')
    redefines_simdutf_implementation = re.compile(r'^#define\s+SIMDUTF_IMPLEMENTATION\s+(.*)')
    undefines_simdutf_implementation = re.compile(r'^#undef\s+SIMDUTF_IMPLEMENTATION\s*$')
    uses_simdutf_implementation = re.compile('SIMDUTF_IMPLEMENTATION([^_a-zA-Z0-9]|$)')
    for line in context.read_file(file):
        s = includepattern.search(line)
        if s:
            includedfile = s.group(2)
            # include all from simdutf.cpp except simdutf.h
            if includedfile == "simdutf.h" and filename == "simdutf.cpp":
                print(line, file=fid)
                continue

            if includedfile.startswith('../'):
                includedfile = includedfile[2:]
            # we explicitly include simdutf headers, one time each (unless they are generic, in which case multiple times is fine)
            doinclude(includedfile, line)
        else:
            # does it contain a redefinition of SIMDUTF_IMPLEMENTATION ?
            s = redefines_simdutf_implementation.search(line)
            if s:
                context.current_implementation = s.group(1)
                print(f"// redefining SIMDUTF_IMPLEMENTATION to \"{context.current_implementation}\"\n// {line}", file=fid)
            elif undefines_simdutf_implementation.search(line):
                # Don't include #undef SIMDUTF_IMPLEMENTATION since we're handling it ourselves
                pass
            else:
                # copy the line, with SIMDUTF_IMPLEMENTATION replace to what it is currently defined to
                print(uses_simdutf_implementation.sub(context.current_implementation+"\\1", line), file=fid)

    print(f"/* end file {RELFILE} */", file=fid)


def get_timestamp():
    # Get the generation date from git, so the output is reproducible.
    # The %ci specifier gives the unambiguous ISO 8601 format, and
    # does not change with locale and timezone at time of generation.
    # Forcing it to be UTC is difficult, because it needs to be portable
    # between gnu date and busybox date.
    try:
        # avoid git going outside simdutf, which could happen when
        # unpacking a release tarball inside a subdirectory of an unrelated
        # git repository. that would lead to picking up the timestamp of the
        # unrelated git repository.
        simdroot = pathlib.Path(SCRIPTPATH).absolute().parent
        GIT_CEILING_DIRECTORIES = str(simdroot.parent)
        ret = subprocess.run(['git', '-C', SCRIPTPATH, 'show', '-s', '--format=%ci', 'HEAD'],
                             stdout=subprocess.PIPE,
                             env=dict(os.environ, GIT_CEILING_DIRECTORIES=GIT_CEILING_DIRECTORIES))

        if ret.returncode != 0:
            print(f"git called resulted in non-zero exit code {ret.returncode}")
            print("timestamp based on current time")
            return str(datetime.datetime.now())

        return ret.stdout.decode('utf-8').strip()
    except (UnicodeDecodeError, FileNotFoundError):
        print("UnicodeDecodeError or FileNotFoundError, timestamp based on current time")
        return str(datetime.datetime.now())


def create_files():
    outdir    = context.args.output_dir
    timestamp = context.timestamp

    os.makedirs(outdir, exist_ok=True)
    AMAL_H = os.path.join(outdir, "simdutf.h")
    AMAL_C = os.path.join(outdir, "simdutf.cpp")

    context.found_includes = set()
    context.current_implementation = ''

    # this list excludes the "src/generic headers"
    ALLCFILES = ["simdutf.cpp"]

    # order matters
    ALLCHEADERS = ["simdutf.h"]

    print(f"Creating {AMAL_H}")
    with open(AMAL_H, 'w') as f:
        context.fid = f
        print(f"/* auto-generated on {timestamp}. Do not edit! */", file=f)
        for header in ALLCHEADERS:
            doinclude(header, f"ERROR {header} not found")

    print(f"Creating {AMAL_C}")
    with open(AMAL_C, 'w') as f:
        context.fid = f
        print(f"/* auto-generated on {timestamp}. Do not edit! */", file=f)
        for cpp in ALLCFILES:
            doinclude(cpp, f"ERROR {cpp} not found")

    # copy the README, DEMO and C API header
    print(f"Copying additional files to {outdir}")
    # C API header
    c_header_path = os.path.join(PROJECTPATH, "include", "simdutf_c.h")
    print(f"Creating {outdir}/simdutf_c.h")
    shutil.copy2(c_header_path, outdir)
    if SCRIPTPATH != outdir:
        for name in ["amalgamation_demo.cpp", "README.md", "amalgamation_demo.c"]:
            print(f"Processing {name}")
            path = os.path.join(SCRIPTPATH, name)
            print(f"Creating {outdir}/{name}")
            shutil.copy2(path, outdir)


def create_zip():
    import zipfile
    outdir = context.args.output_dir

    path = os.path.join(outdir, context.zipname)
    print(f"Creating {path}")
    with zipfile.ZipFile(path, 'w') as zf:
        for name in ["simdutf.cpp", "simdutf.h", "amalgamation_demo.cpp", "README.md", "simdutf_c.h", "amalgamation_demo.c"]:
            source = os.path.join(outdir, name)
            zf.write(source, name)


def print_instructions():
    README = os.path.join(context.args.output_dir, "README.md")
    print()
    print("Giving final instructions:")
    print()
    sys.stdout.write(open(README).read())
    print()


def read_file(file):
    with open(file, 'r') as f:
        for line in f:
            yield line.rstrip()


def filter_features(file):
    """
    Design:

    * Feature macros SIMDUTF_FEATURE_foo must not be nested.
    * All #endifs must contain a comment with the repeated condition.
    """
    current_features = None
    start_if_line = None
    enabled = True
    prev_line = ''

    root_header = file.endswith("/implementation.h")

    with open(file, 'r') as f:
        lines = [line.rstrip() for line in f.readlines()]
        for (lineno, line) in enumerate(lines, 1):
            if line is None:
                continue

            if root_header and line.startswith('#define SIMDUTF_FEATURE'):
                # '#define SIMDUTF_FEATURE_FOO 1'
                tmp = line.split()
                assert len(tmp) == 3, line
                assert tmp[2] == '1'
                flag = tmp[1]

                if flag in context.enabled_features:
                    yield line
                else:
                    yield f'#define {flag} 0'

            elif line.startswith('#if SIMDUTF_FEATURE'):
                if start_if_line is not None:
                    raise ValueError(f"{file}:{lineno}: feature block already opened at line {start_if_line}")

                prefix_len = len('#if ')
                if line.endswith('\\'):
                    nextline = lines[lineno]
                    lines[lineno] = None

                    expr = line[prefix_len:-1] + nextline
                else:
                    expr = line[prefix_len:]

                current_features = get_features(file, lineno, expr)
                start_if_line = lineno
                enabled = current_features.evaluate(context.enabled_features)
            elif line.startswith('#endif') and line.find('// SIMDUTF_FEATURE') > 0:
                if start_if_line is None:
                    raise ValueError(f"{file}:{lineno}: feature block not opened, orphan #endif found")

                _, _, expr = line.partition('//')
                if lineno < len(lines):
                    nextline = lines[lineno].lstrip()
                    if nextline.startswith('//') and nextline.find('SIMDUTF_FEATURE') > 0:
                        expr += nextline[len('//'):]
                        lines[lineno] = None

                features = get_features(file, lineno, expr)
                if str(features) != str(current_features):
                    raise ValueError(f"{file}:{lineno}: feature #endif condition different than opening #if at {start_if_line}")

                enabled = True
                start_if_line = None
                current_features = None
            elif enabled:
                if context.args.debug_sources and not prev_line.endswith('\\'):
                    RELFILE = os.path.relpath(file, PROJECTPATH)
                    yield f"// {RELFILE}:{lineno}"

                if line or (not line and prev_line):
                    yield line

                prev_line = line


def get_features(file, lineno, line):
    try:
        return parse_condition(line)
    except Exception as e:
        raise ValueError(f"{file}:{lineno}: {e}")


class Token:
    def __init__(self, name):
        if name not in known_features:
            raise ValueError(f"unknown feature name '{name}'")

        self.name = name

    def evaluate(self, enabled_features):
        return self.name in enabled_features

    def __str__(self):
        return self.name


class And:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def evaluate(self, enabled_features):
        a = self.a.evaluate(enabled_features)
        b = self.b.evaluate(enabled_features)

        return a and b

    def __str__(self):
        return '(%s && %s)' % (self.a, self.b)


class Or:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def evaluate(self, enabled_features):
        a = self.a.evaluate(enabled_features)
        b = self.b.evaluate(enabled_features)

        return a or b

    def __str__(self):
        return '(%s || %s)' % (self.a, self.b)


def parse_condition(s):
    tokens = [t for t in re.split('( |\\(|\\)|&&|\\|\\|)', s) if t not in ('', ' ')]
    stack  = []

    expr = None
    for token in tokens:
        if token == '&&':
            stack.append(token)
        elif token == '||':
            stack.append(token)
        elif token == '(':
            stack.append(expr)
            expr = None
        elif token == ')':
            prev = stack.pop()
            op   = stack.pop()
            if op == '&&':
                expr = And(prev, expr)
            elif op == '||':
                expr = Or(prev, expr)
            else:
                assert False, op
        else:
            e = Token(token)
            if expr is None:
                expr = e
            else:
                assert stack, "empty expression stack"
                op = stack.pop()
                if op == '&&':
                    expr = And(expr, e)
                elif op == '||':
                    expr = Or(expr, e)
                else:
                    assert False, op

    return expr


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
from pathlib import Path
from io import StringIO
import os


TMP = Path("/dev/shm/")
if not TMP.exists():
    TMP = Path(os.environ.get('TMP', '.'))


def main():
    compilers = [
        ('default', 'c++', [])
    ]

    crosscompilers = find_crosscompilers()
    if crosscompilers:
        print("Found the following crosscompilers in $PATH:")
        for arch, compiler, opts in crosscompilers:
            print('%-16s: %s' % (arch, compiler))

        compilers.extend(crosscompilers)

    make = create_make(compilers)

    file = Path('Makefile')
    update_file(make, file)


def create_make(compilers):
    f = StringIO()

    def writeln(s):
        f.write(s + '\n')

    writeln("SRC=../src")
    writeln("INC=../include")

    archs = []
    for arch, compiler, opts in compilers:
        targets = []
        for features in feature_combinations:
            parts = [arch] + name_for_features(features)

            target_dir = TMP / '-'.join(parts)
            target_obj = target_dir / 'simdutf.o'
            target_exe = target_dir / 'a.out'

            targets.append((target_obj, target_exe, features))

        archs.append((arch, compiler, opts, targets))

        writeln('')
        writeln(f"{arch.upper()}=\\")
        for (target_obj, target_exe, _) in targets:
            writeln(f'\t{target_exe} {target_obj}\\')

    # first add 'all' target
    writeln('')
    writeln('.PHONY: all')
    writeln('all: %s' % (' '.join(arch for arch, _, _ in compilers)))
    writeln('')
    writeln('.PHONY: all')
    writeln('clean:')
    writeln('\t$(RM) %s' % ' '.join(f"$({arch.upper()})" for arch, _, _ in compilers))

    # add 'help' target
    writeln('')
    writeln(".PHONY: help")
    writeln("help:")
    width = max((len(arch) for arch, _, _ in compilers))
    width = max(width, len("all"), len("clear"))

    target = "all"
    writeln(f'\t@echo "make {target:{width}} --- build all targets"')
    target = "clear"
    writeln(f'\t@echo "make {target:{width}} --- remove all generated files"')
    target = "help"
    writeln(f'\t@echo "make {target:{width}} --- show this help"')
    writeln('\t@echo')
    for arch, compiler, _ in compilers:
        writeln(f'\t@echo "make {arch:{width}} --- build using {compiler}"')

    # add convenient targets for architectures ('default' is the first)
    for arch, _, _ in compilers:
        writeln('')
        writeln(f'.PHONY: {arch}')
        writeln(f'{arch}: $({arch.upper()})')

    # add individual targets
    for arch, compiler, opts, targets in archs:
        compiler_opts   = ' '.join(opts)
        for target_obj, target_exe, features in targets:
            amalgamate_opts = ' '.join(feature2option[feat] for feat in features)

            target_dir = target_obj.parent

            writeln('')
            writeln(f"{target_obj}: amalgamate.py")
            writeln(f"\tmkdir -p {target_dir}")
            writeln(f"\tpython3 amalgamate.py --no-zip --no-readme --source-dir=$(SRC) --include-dir=$(INC) --output {target_dir} {amalgamate_opts}")
            writeln(f"\tcd {target_dir} && {compiler} {compiler_opts} -c simdutf.cpp")
            writeln('')
            writeln(f"{target_exe}: {target_obj}")
            writeln(f"\tcd {target_dir} && {compiler} {compiler_opts} amalgamation_demo.cpp")

    return f.getvalue()


def update_file(contents, path):
    if path.exists():
        if path.read_text() == contents:
            return

        print(f"updating {path}")
    else:
        print(f"creating {path}")

    path.write_text(contents)


def find_crosscompilers():
    return list(find_crosscompilers_aux())


def find_crosscompilers_aux():
    found = set()
    for path in os.environ['PATH'].split(':'):
        path = Path(path)
        gxx = glob_many(path, ['*-g++*', '*-c++*', '*clang++'])
        for item in crosscompilers:
            if isinstance(item, str):
                arch = item
                name = arch
                opts = []
            elif isinstance(item, tuple):
                assert len(item) == 3, item
                arch, name, opts = item
            else:
                assert False, item

            if name in found:
                continue

            for filename in gxx:
                if is_compiler(arch, ['g++', 'c++', 'clang++'], filename):
                    yield (name, filename, opts)
                    found.add(name)
                    break


def glob_many(rootdir, patterns):
    tmp = []
    for pat in patterns:
        tmp.extend([file.name for file in rootdir.glob(pat)])

    tmp.sort()
    return tmp


def is_compiler(arch, compilers, name):
    # we're looking for "arch-foo-bar-g++" or "arch-foo-bar-g++-version"
    tmp = name.split('-')
    if tmp[0] != arch:
        return False

    if tmp[-1] in compilers:
        return True

    if len(tmp) >= 3 and tmp[-2] in compilers and is_number(tmp[-1]):
        return True


def is_number(s):
    try:
        _ = int(s)
        return True
    except ValueError:
        return False


def name_for_features(features):
    return [feature2stem[feat] for feat in features]


SIMDUTF_FEATURE_DETECT_ENCODING = 'SIMDUTF_FEATURE_DETECT_ENCODING'
SIMDUTF_FEATURE_LATIN1          = 'SIMDUTF_FEATURE_LATIN1'
SIMDUTF_FEATURE_ASCII           = 'SIMDUTF_FEATURE_ASCII'
SIMDUTF_FEATURE_BASE64          = 'SIMDUTF_FEATURE_BASE64'
SIMDUTF_FEATURE_UTF8            = 'SIMDUTF_FEATURE_UTF8'
SIMDUTF_FEATURE_UTF16           = 'SIMDUTF_FEATURE_UTF16'
SIMDUTF_FEATURE_UTF32           = 'SIMDUTF_FEATURE_UTF32'


feature2stem = {
    SIMDUTF_FEATURE_DETECT_ENCODING : 'de',
    SIMDUTF_FEATURE_LATIN1          : 'lat1',
    SIMDUTF_FEATURE_ASCII           : 'ascii',
    SIMDUTF_FEATURE_BASE64          : 'base64',
    SIMDUTF_FEATURE_UTF8            : 'utf8',
    SIMDUTF_FEATURE_UTF16           : 'utf16',
    SIMDUTF_FEATURE_UTF32           : 'utf32',
}

feature2option = {
    SIMDUTF_FEATURE_DETECT_ENCODING : '--with-detect-enc',
    SIMDUTF_FEATURE_LATIN1          : '--with-latin1',
    SIMDUTF_FEATURE_ASCII           : '--with-ascii',
    SIMDUTF_FEATURE_BASE64          : '--with-base64',
    SIMDUTF_FEATURE_UTF8            : '--with-utf8',
    SIMDUTF_FEATURE_UTF16           : '--with-utf16',
    SIMDUTF_FEATURE_UTF32           : '--with-utf32',
}

feature_combinations = [
    [SIMDUTF_FEATURE_DETECT_ENCODING],
    [SIMDUTF_FEATURE_ASCII],
    [SIMDUTF_FEATURE_UTF8],
    [SIMDUTF_FEATURE_UTF16],
    [SIMDUTF_FEATURE_UTF32],
    [SIMDUTF_FEATURE_BASE64],
    [SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING],
    [SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING],
    [SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII],
    [SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_LATIN1, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII],
    [SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_UTF16],
    [SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF32],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF8],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF16],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF8],
    [SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_UTF32, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_LATIN1],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF8],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF16],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF32],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_UTF16],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_UTF32],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF32],
    [SIMDUTF_FEATURE_BASE64, SIMDUTF_FEATURE_DETECT_ENCODING, SIMDUTF_FEATURE_ASCII, SIMDUTF_FEATURE_UTF8, SIMDUTF_FEATURE_UTF16, SIMDUTF_FEATURE_UTF32],
]

crosscompilers = [
    'aarch64',
    'powerpc64',
    'loongarch64',
    ('loongarch64', 'loongarch64lasx', ["-mlsx", "-mlasx"]),
    ('riscv64', 'riscv64', ["-march=rv64gv"]),
]

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import subprocess
import glob
import pathlib
import shutil
import re
import base64


def generate_readme():
    # we rely on html comments in the markdown to process the file into
    # what we desire in the github pages content, generated by doxygen.
    # this is very simple and brittle.
    marker = re.compile(
        r"^<!-- (.*) \(this is used by the github pages export, don't modify!\) -->$")
    with open('README.md') as input, open('README.generated.md', 'w') as output:
        githubonly = False
        for line in input:
            if m := marker.match(line):
                directive = m.group(1).strip()
                print(f"got match {directive}")
                if directive == 'if(github) {':
                    assert not githubonly
                    githubonly = True
                elif directive == '}':
                    assert githubonly
                    githubonly = False
                elif directive.startswith('base64 '):
                    print(base64.b64decode(directive[7:]).decode(
                        "utf-8"), file=output)
                else:
                    raise Exception('failed to parse directive')

            if not githubonly:
                print(line, file=output)


def main():
    subprocess.run(['./scripts/prepare_doxygen.sh',])
    generate_readme()
    subprocess.run(['doxygen',])
    pathlib.Path('doc/api/html/doc').mkdir(parents=True, exist_ok=True)
    for file in glob.glob('doc/*png'):
        print(f"copying {file}")
        shutil.copy(file, 'doc/api/html/doc')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Script to automate adding a new function to the simdutf library.
This script reads a file containing the function signatures and documentation,
then adds the functions to the appropriate files in the codebase.

Usage:
    python add_function.py <signature_file>

Where <signature_file> is a file containing the function signatures, wrapped in a feature macro block, e.g.:
    #if SIMDUTF_FEATURE_UTF8 && SIMDUTF_FEATURE_UTF16
    /**
     * Documentation for first function.
     */
    simdutf_warn_unused size_t utf8_length_from_utf16le(
        const char16_t *buf, size_t len) noexcept;

    /**
     * Documentation for second function.
     */
    simdutf_warn_unused size_t utf8_length_from_utf16be(
        const char16_t *buf, size_t len) noexcept;
    #endif
"""

import sys
import os
import re
import glob

def read_signature_file(file_path):
    """Read and parse the signature file."""
    with open(file_path, 'r') as f:
        content = f.read()
    
    # Extract the feature macro (e.g., SIMDUTF_FEATURE_UTF8 && SIMDUTF_FEATURE_UTF16)
    feature_match = re.search(r'#if\s+(.+)', content)
    feature_macro = feature_match.group(1).strip() if feature_match else None
    
    # Extract the content between #if and #endif
    if_match = re.search(r'#if\s+.+?(#endif)', content, re.DOTALL)
    if not if_match:
        raise ValueError("Could not find #if/#endif block.")
    
    block_content = if_match.group(0)
    
    # Extract all function signatures (documentation + signature)
    sig_matches = re.findall(r'(/\*\*.*?\*/)\s*(.*?);', block_content, re.DOTALL)
    if not sig_matches:
        raise ValueError("Could not parse function signatures from file.")
    
    functions = []
    for doc, sig in sig_matches:
        doc = doc.strip()
        signature = sig.strip() + ';'
        # Extract function name
        name_match = re.search(r'\s+(\w+)\s*\(', signature)
        func_name = name_match.group(1) if name_match else None
        functions.append((doc, signature, func_name))
    
    return feature_macro, functions

def add_to_implementation_h(repo_root, feature_macro, functions):
    """Add to include/simdutf/implementation.h"""
    file_path = os.path.join(repo_root, 'include/simdutf/implementation.h')
    
    with open(file_path, 'r') as f:
        content = f.read()
    
    # Check for existing #if block for standalone functions (after namespace)
    ns_start = content.find('namespace simdutf {')
    if ns_start == -1:
        raise ValueError("Could not find namespace in implementation.h")
    
    existing_if = content.find(f'#if {feature_macro}', ns_start)
    if existing_if != -1:
        # Find the corresponding #endif
        endif_pos = content.find('#endif', existing_if)
        if endif_pos != -1:
            insert_pos = endif_pos
        else:
            insert_pos = len(content)
    else:
        insert_pos = ns_start + len('namespace simdutf {')
    
    # Insert standalone functions
    standalone = ""
    for doc, signature, func_name in functions:
        standalone_sig = signature.replace(' const ', ' ')
        standalone += f"""{doc}

{standalone_sig}


"""
    content = content[:insert_pos] + standalone + content[insert_pos:]
    
    # Now for virtual functions in class
    class_start = content.find('class implementation {')
    public_start = content.find('public:', class_start)
    if public_start == -1:
        raise ValueError("Could not find public section in implementation class")
    
    existing_if_class = content.find(f'#if {feature_macro}', public_start)
    if existing_if_class != -1:
        endif_pos_class = content.find('#endif', existing_if_class)
        if endif_pos_class != -1:
            insert_pos_class = endif_pos_class
        else:
            insert_pos_class = len(content)
    else:
        insert_pos_class = public_start + len('public:')
    
    # Insert virtual functions
    virtual = ""
    for doc, signature, func_name in functions:
        # Normalize signature for virtual
        clean_sig = re.sub(r'\s*;?\s*$', '', signature).strip()
        virtual += f"""{doc}
  virtual {clean_sig} = 0;


"""
    content = content[:insert_pos_class] + virtual + content[insert_pos_class:]
    
    with open(file_path, 'w') as f:
        f.write(content)

def add_to_src_implementation_cpp(repo_root, feature_macro, functions):
    """Add to src/implementation.cpp"""
    file_path = os.path.join(repo_root, 'src/implementation.cpp')
    
    with open(file_path, 'r') as f:
        content = f.read()
    
    # For detect class
    detect_start = content.find('class detect_best_supported_implementation_on_first_use')
    if detect_start != -1:
        existing_if = content.find(f'#if {feature_macro}', detect_start)
        if existing_if != -1:
            endif_pos = content.find('#endif', existing_if)
            insert_pos = endif_pos if endif_pos != -1 else len(content)
        else:
            brace_pos = content.find('{', detect_start)
            insert_pos = brace_pos + 1 if brace_pos != -1 else len(content)
    else:
        insert_pos = len(content)
    
    detect_impl = ""
    for doc, signature, func_name in functions:
        params = [p.strip().split()[-1].lstrip('*') for p in signature.split('(')[1].split(')')[0].split(',') if p.strip()]
        param_list = ', '.join(params)
        detect_impl += f"""{re.sub(r';$', ' {', signature.replace(' const noexcept', ' const noexcept final override'))}
  return set_best()->{func_name}({param_list});
}}


"""
    content = content[:insert_pos] + detect_impl + content[insert_pos:]
    
    # For unsupported class
    unsupported_start = content.find('class unsupported_implementation final : public implementation {')
    if unsupported_start != -1:
        existing_if = content.find(f'#if {feature_macro}', unsupported_start)
        if existing_if != -1:
            endif_pos = content.find('#endif', existing_if)
            insert_pos_unsup = endif_pos if endif_pos != -1 else len(content)
        else:
            insert_pos_unsup = unsupported_start + len('class unsupported_implementation final : public implementation {')
    else:
        insert_pos_unsup = len(content)
    
    unsupported_impl = ""
    for doc, signature, func_name in functions:
        return_stmt = "return nullptr;" if "const char" in signature.split('(')[0] or "char16_t" in signature.split('(')[0] else "return 0;"
        unsupported_impl += f"""{re.sub(r';$', ' {', signature.replace(' const noexcept', ' const noexcept final override'))}
  {return_stmt}  // Not supported
}}


"""
    content = content[:insert_pos_unsup] + unsupported_impl + content[insert_pos_unsup:]
    
    # For standalone functions at end
    existing_if_end = content.rfind(f'#if {feature_macro}')
    if existing_if_end != -1:
        endif_pos = content.find('#endif', existing_if_end)
        insert_pos_end = endif_pos if endif_pos != -1 else len(content)
    else:
        insert_pos_end = len(content)
    
    standalone = ""
    for doc, signature, func_name in functions:
        params = [p.strip().split()[-1].lstrip('*') for p in signature.split('(')[1].split(')')[0].split(',') if p.strip()]
        param_list = ', '.join(params)
        # Remove const for standalone functions
        standalone_sig = signature.replace(' const noexcept', ' noexcept')
        standalone += f"""{re.sub(r';$', ' {', standalone_sig)}
  return get_default_implementation()->{func_name}({param_list});
}}


"""
    content = content[:insert_pos_end] + standalone + content[insert_pos_end:]
    
    with open(file_path, 'w') as f:
        f.write(content)

def add_to_impl_files(file_path, functions, feature_macro):
    with open(file_path, 'r') as f:
        content = f.read()
    
    impl = ""
    for doc, signature, func_name in functions:
        modified_signature = re.sub(re.escape(func_name), f'implementation::{func_name}', signature)
        if ' const ' not in modified_signature:
            modified_signature = modified_signature.replace(' noexcept', ' const noexcept')
        modified_signature = modified_signature.replace(' const noexcept', ' const noexcept final override')
        impl += f"""{re.sub(r';$', ' {', modified_signature)}
  // TODO: implement
}}


"""
    
    existing_if = content.rfind(f'#if {feature_macro}')
    if existing_if != -1:
        endif_pos = content.find('#endif', existing_if)
        insert_pos = endif_pos if endif_pos != -1 else len(content)
        content = content[:insert_pos] + impl + content[insert_pos:]
    else:
        impl = f"#if {feature_macro}\n{impl}#endif // {feature_macro}\n"
        content += impl
    
    with open(file_path, 'w') as f:
        f.write(content)

def add_declarations_to_impl_h(file_path, functions, feature_macro):
    print(f"Adding declarations to {file_path}...")
    # Now add declarations to the corresponding .h file
    arch = os.path.basename(os.path.dirname(file_path))
    h_file = file_path
    with open(h_file, 'r') as f:
        h_content = f.read()
        
    
    insert_pos = h_content.rfind('};')
    if insert_pos == -1:
        print(f"}}; in {h_file}")
    if insert_pos != -1:
        print(f"Inserting declarations at position {insert_pos}...")
        decl = ""
        for doc, signature, func_name in functions:
            modified_signature = signature
            if ' const ' not in modified_signature:
                modified_signature = modified_signature.replace(' noexcept', ' const noexcept')
            decl += f"{modified_signature};\n"
        
        # Wrap with #if
        existing_if_h = h_content.rfind(f'#if {feature_macro}')
        if existing_if_h != -1:
            endif_pos_h = h_content.find('#endif', existing_if_h)
            insert_pos_h = endif_pos_h if endif_pos_h != -1 else insert_pos
            h_content = h_content[:insert_pos_h] + decl + h_content[insert_pos_h:]
        else:
            decl = f"#if {feature_macro}\n{decl}#endif // {feature_macro}\n"
            h_content = h_content[:insert_pos] + decl + h_content[insert_pos:]
        print(f"Updating {h_file}...")
        with open(h_file, 'w') as f:
            f.write(h_content)

def add_to_all_impl_files(repo_root, feature_macro, functions):
    """Add to all src/**/implementation.cpp files"""
    impl_files = glob.glob(os.path.join(repo_root, 'src', '**', 'implementation.cpp'), recursive=True)
    # Exclude src/implementation.cpp as it's handled separately
    impl_files = [f for f in impl_files if os.path.dirname(f) != os.path.join(repo_root, 'src')]
    print(f"Found {len(impl_files)} implementation.cpp files to update.")
    
    for file_path in impl_files:
        print(f"Updating {file_path}...")
        add_to_impl_files(file_path, functions, feature_macro)

def add_declaration_to_all_impl_files(repo_root, feature_macro, functions):
    """Add declarations to all src/simdutf/**/implementation.h files"""
    impl_files = glob.glob(os.path.join(repo_root, 'src/simdutf', '**', 'implementation.h'), recursive=True)
    # Exclude src/implementation.cpp as it's handled separately
    impl_files = [f for f in impl_files if os.path.dirname(f) != os.path.join(repo_root, 'src')]
    print(f"Found {len(impl_files)} implementation.h files to update.")
    
    for file_path in impl_files:
        print(f"Updating declarations for {file_path}...")
        add_declarations_to_impl_h(file_path, functions, feature_macro)

def main():
    if len(sys.argv) != 2:
        print("Usage: python add_function.py <signature_file>")
        sys.exit(1)
    
    script_dir = os.path.dirname(os.path.abspath(__file__))
    repo_root = os.path.abspath(os.path.join(script_dir, '..'))
    
    sig_file = sys.argv[1]
    feature_macro, functions = read_signature_file(sig_file)
    
    # Ensure all signatures have const
    updated_functions = []
    for doc, sig, name in functions:
        if ' const ' not in sig:
            sig = sig.replace(' noexcept', ' const noexcept')
        updated_functions.append((doc, sig, name))
    functions = updated_functions
    
    if not feature_macro or not functions:
        print("Error: Could not extract feature macro or functions.")
        sys.exit(1)
    
    #add_to_implementation_h(repo_root, feature_macro, functions)
    #add_to_src_implementation_cpp(repo_root, feature_macro, functions)
    #add_to_all_impl_files(repo_root, feature_macro, functions)
    add_declaration_to_all_impl_files(repo_root, feature_macro, functions)
    
    func_names = [name for _, _, name in functions]
    print(f"Functions '{', '.join(func_names)}' added successfully.")
    print("Please remember to implement the functions in the respective implementation.cpp files.")
    print("Also, consider adding tests for the new functions.")
    print("Run formatting tools to ensure code style consistency: ./scripts/clang_format_docker.sh ")
    print("Done.")

if __name__ == '__main__':
    main()


#!/usr/bin/env python3

import sys
from pathlib import Path
from table import Table


class Input:
    def __init__(self):
        self.procedure = None
        self.input_size = None
        self.iterations = None
        self.dataset = None

    @property
    def dataset_name(self):
        if self.dataset:
            return self.dataset.stem
        else:
            return 'none'

    def __str__(self):
        return '<Input: %s, size: %d, iterations: %s, dataset: %s>' % \
                (self.procedure, self.input_size, self.iterations, self.dataset)

    __repr__ = __str__


class Result:
    def __init__(self):
        self.instruction_per_byte = None
        self.instruction_per_char = None
        self.instruction_per_cycle = None
        self.cycle_per_byte = None
        self.cycle_per_char = None
        self.bytes_per_char = None
        self.speed_gigabytes = None
        self.speed_gigachars = None
        self.freq = None
        self.time = None

    def update_from_dict(self, D):
        while D:
            unit, value = D.popitem()
            field = unit2field[unit]
            if field is not None:
                setattr(self, field, value)


unit2field = {
    'ins/byte'      : 'instruction_per_byte',
    'ins/char'      : 'instruction_per_char',
    'ins/cycle'     : 'instruction_per_cycle',
    'cycle/byte'    : 'cycle_per_byte',
    'cycle/char'    : 'cycle_per_char',
    'byte/char'     : 'byte_per_char',
    'GHz'           : 'freq',
    'GB/s'          : 'speed_gigabytes',
    'Gc/s'          : 'speed_gigachars',
    'ns'            : 'time',
    '%'             : None,
}


def parse(file):
    result = []
    for line in file:
        for item in parse_line(line):
            if isinstance(item, Input):
                result.append([item, None])
            else:
                assert isinstance(item, dict)
                if result[-1][1] is None:
                    result[-1][1] = Result()

                result[-1][1].update_from_dict(item)

    return [(input, result) for (input, result) in result if result is not None]


def parse_line(line):
    if 'input size' in line:
        yield parse_input(normalize_line(line))
    elif 'GB/s' in line or 'ins/byte' in line or 'ins/char' in line:
        yield parse_results(normalize_line(line))


def normalize_line(line):
    line = line.replace(',', ' ')
    line = line.replace(':', ' ')
    line = line.replace('(', ' ')
    line = line.replace(')', ' ')

    return line.split()


def parse_input(fields):
    input = Input()
    input.procedure = fields.pop(0)

    assert fields.pop(0) == 'input'
    assert fields.pop(0) == 'size'
    input.input_size = int(fields.pop(0))

    assert fields.pop(0) == 'iterations'
    input.iterations = int(fields.pop(0))

    try:
        assert fields.pop(0) == 'dataset'
        input.dataset = Path(fields.pop(0))
    except IndexError:
        pass

    return input


def try_float(s):
    try:
        return float(s)
    except ValueError:
        return


def parse_results(fields):
    """Consumes pairs "number unit" from the list"""
    D = {}

    while fields:
        value = try_float(fields.pop(0))
        if value is not None and fields:
            D[fields.pop(0)] = value

    return D


def speed_comparison(data):
    procedures = set()
    datasets = set()
    results = {}

    for input, result in data:
        procedures.add(input.procedure)
        datasets.add(input.dataset_name)
        results[(input.procedure, input.dataset_name)] = result

    datasets = list(sorted(datasets))
    procedures = list(sorted(procedures))

    def by_procedure():
        table = Table()
        table.set_header(['procedure'] + datasets)

        for procedure in procedures:
            row = []
            row.append(procedure)
            for dataset in datasets:
                try:
                    result = results[(procedure, dataset)]
                    row.append('%0.3f GB/s' % (result.speed_gbs))
                except KeyError:
                    row.append('--')

            table.add_row(row)

        return table

    def by_dataset():
        table = Table()
        table.set_header(['dataset'] + procedures)

        for dataset in datasets:
            row = []
            row.append(dataset)
            for procedure in procedures:
                try:
                    result = results[(procedure, dataset)]
                    row.append('%0.3f GB/s' % (result.speed_gigabytes))
                except KeyError:
                    row.append('--')

            table.add_row(row)

        return table

    if len(procedures) >= len(datasets):
        return by_procedure()
    else:
        return by_dataset()


def compare(old, new):
    collate = {}
    for input, result in old:
        key = str(input)
        collate[key] = [input, result, None]

    for input, result in new:
        key = str(input)
        if key not in collate:
            collate[key] = [input, None, result]
        else:
            collate[key][-1] = result

    grouped = {}
    for _, (input, old, new) in collate.items():
        if input.procedure not in grouped:
            grouped[input.procedure] = []

        grouped[input.procedure].append((input, old, new))

    NA = '--'
    table = Table()
    table.set_header(["dataset", "size [B]", "iterations", "old GB/s", "new GB/s", "speedup"])
    for procedure, values in grouped.items():
        table.add_row([(procedure, 6)])
        for input, old, new in values:
            old_speed = old.speed_gigabytes
            new_speed = new.speed_gigabytes
            if old_speed is None or new_speed is None:
                speedup = NA
            else:
                speedup = '%0.2fx' % (new_speed / old_speed)

            if old_speed is None:
                old_speed = NA
            else:
                old_speed = '%0.2f' % old_speed

            if new_speed is None:
                new_speed = NA
            else:
                new_speed = '%0.2f' % new_speed

            table.add_row([
                input.dataset.stem,
                '%s' % input.input_size,
                '%s' % input.iterations,
                old_speed,
                new_speed,
                speedup,
            ])

    return table


HELP = """Format output of the benchmark utility

When running a benchmark redirect its output to file (with > or `tee`), like:

$ cd build/benchmarks
$ ./benchmark [...] > results.txt
$ ./benchmark [...] | tee results.txt

To summarize results, use:

$ ./benchmark_print.py results.txt
|     dataset      | utf8_length_from_utf32+fallback | utf8_length_from_utf32+haswell | utf8_length_from_utf32+westmere |
| ---------------- | ------------------------------- | ------------------------------ | ------------------------------- |
| arabic.utf32     | 3.190 GB/s                      | 15.638 GB/s                    | 7.879 GB/s                      |
| chinese.utf32    | 3.179 GB/s                      | 15.349 GB/s                    | 7.644 GB/s                      |
| czech.utf32      | 3.178 GB/s                      | 15.721 GB/s                    | 7.801 GB/s                      |
...
| thai.utf32       | 3.191 GB/s                      | 15.455 GB/s                    | 7.842 GB/s                      |
| turkish.utf32    | 3.191 GB/s                      | 15.086 GB/s                    | 7.809 GB/s                      |
| vietnamese.utf32 | 3.182 GB/s                      | 15.317 GB/s                    | 7.807 GB/s                      |

To compare results, use:

$ ./benchmark_print.py old.txt new.txt
|             dataset             | size [B] | iterations | old GB/s | new GB/s | speedup |
| ------------------------------- | -------- | ---------- | -------- | -------- | ------- |
| convert_utf32_to_latin1+haswell                                                         |
| esperanto.utflatin32            |   328672 |      30000 |    77.14 |    88.81 | 1.15x   |
| french.utflatin32               |  1729220 |      30000 |    75.26 |    92.39 | 1.23x   |
| german.utflatin32               |   797324 |      30000 |    76.11 |    79.84 | 1.05x   |
| portuguese.utflatin32           |  1086972 |      30000 |    84.12 |    92.64 | 1.10x   |
"""


def parse_file(path):
    with open(path, 'r') as f:
        return parse(f)


def self_test():
    rootdir = Path(__file__).parent / 'tests'

    data = parse_file(rootdir / 'results.txt')
    speed_comparison(data)

    old = parse_file(rootdir / 'old.txt')
    new = parse_file(rootdir / 'old.txt')
    compare(new, old)


def main():
    args = sys.argv[1:]
    if "-h" in args or "--help" in args:
        print(HELP)
        return

    if "--test" in args:
        print("running tests...")
        self_test()
        print("all OK")
        return

    n = len(args)

    if n == 1:
        data = parse_file(args[0])

        print(speed_comparison(data))
    elif n == 2:
        old = parse_file(args[0])
        new = parse_file(args[1])

        print(compare(old, new))
    else:
        print(HELP)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# runs a compilation benchmark

import subprocess
import tempfile
import argparse
import datetime
import statistics


class Sample:
    def __init__(self, t0, t1, t2):
        self._configure = (t1-t0).total_seconds()
        self._compile = (t2-t1).total_seconds()

# benchmarks compilation, returns the time consumption


def measure_once(standard: int):
    builddir = tempfile.TemporaryDirectory()
    t0 = datetime.datetime.now()
    subprocess.run(["cmake", f"-DSIMDUTF_CXX_STANDARD={standard}", "-DSIMDUTF_FAST_TESTS=On",
                   "-B", builddir.name, "-GNinja"], stdout=subprocess.DEVNULL)
    t1 = datetime.datetime.now()
    subprocess.run(["cmake", "--build", builddir.name],
                   stdout=subprocess.DEVNULL)
    t2 = datetime.datetime.now()
    # print(f"configure took {t1-t0}, build took {t2-t1}")
    return Sample(t0, t1, t2)


def benchmark_single(standard: int, repetitions: int = 3):
    d = [measure_once(standard) for i in range(repetitions)]

    m = statistics.median([e._configure for e in d])
    print(f"median configure is {m:.3f}s")

    m = statistics.median([e._compile for e in d])
    print(f"median compile is {m:.3f}s")


def main():
    parser = argparse.ArgumentParser(
        prog='compilation benchmark',
        description='benchmarks compilation of simdutf')
    parser.add_argument('--standard', type=int, default=23)
    parser.add_argument('--repetitions', type=int, default=3)

    args = parser.parse_args()
    benchmark_single(standard=args.standard, repetitions=args.repetitions)


if __name__ == '__main__':
    main()



import sys
delta_asso = [0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00,0x00, 0x00, 0x00, 0x0F, 0x00, 0x0F]
check_asso = [0x0D, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x03, 0x07, 0x0B, 0x0B, 0x0B, 0x0F]

delta_values =[(0x00), (0x00), (0x00), (0x13), (0x04), (0xBF), (0xBF), (0xB9), (0xB9), (0x00), (0x10), (0xC3), (0xBF), (0xBF), (0xB9), (0xB9)]
check_values = [(0x80), (0x80), (0x80), (0x80), (0xCF), (0xBF), (0xD5), (0xA6), (0xB5), (0x86), (0xD1), (0x80), (0xB1), (0x80), (0x91), (0x80)]


valid = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"

def safechr(i):
    if i < 32:
        return '.'
    if i > 127:
        return '?'
    return chr(i)

def safehex(x):
    return "0x{0:2x}".format(x)

def to_signed(x):
    if(x >= 128):
        return x - 256
    return x

def to_unsigned(x):
    if(x < 0):
        return x + 256
    return x

def sat(x, y):
    x = to_signed(x)
    y = to_signed(y)
    z = x + y
    if(z > 127):
        return 127
    if(z < -128):
        return to_unsigned(-128)
    return to_unsigned(z)

def lookup(table, index):
    print("looking up ", hex(index))
    if(index >= 128):
        return 0
    return table[index&0xf]



def quietlookup(table, index):
    if(index >= 128):
        return 0
    return table[index&0xf]

def process(src):
    shifted = (src >> 3)%256
    delta_hash = (lookup(delta_asso,src) + shifted + 1) >> 1
    check_hash = (lookup(check_asso,src) + shifted + 1) >> 1
    out = sat(lookup(delta_values,delta_hash), src)
    chk = sat(lookup(check_values,check_hash), src)
    mask = chk & 0x80
    return (out, mask)

def processquiet(src):
    shifted = (src >> 3)%256
    delta_hash = (quietlookup(delta_asso,src) + shifted + 1) >> 1
    check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
    out = sat(quietlookup(delta_values,delta_hash), src)
    chk = sat(quietlookup(check_values,check_hash), src)
    mask = chk & 0x80
    return (out, mask)

def is_ok(i):
    out, mask = processquiet(i)
    if mask == 0:
        return 1
    return 0

def computestring():
    s = ""
    for i in range(256):
        out, mask = processquiet(i)
        if(mask == 0):
            s +=  safechr(i)
    return s
#print(computestring() + " " + str(len(computestring())))

def print_layout():
    t={}
    for i in range(256):
        src = i
        shifted = (src >> 3)%256
        check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
        if check_hash not in t:
            t[check_hash] = []
        t[check_hash].append(i)
    for check_hash in range(16):
        if check_hash in t:
            off = quietlookup(check_values,check_hash)
            print(hex(check_hash), hex(off), end="")
            print("\t", " ".join(["   "+safechr(c) for c in t[check_hash]]))
        else:
            continue


def is_valid():
    t={}
    # we compute the hash for each character, we store the character in a list for each hash value
    for i in range(256):
        src = i
        for j in range(8):
            shifted = (src >> 3)%256
            shifted += j << 5
            check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
            if check_hash not in t:
                t[check_hash] = []
            t[check_hash].append(i)
    # for each hash value, we check if the content of the sequence. It should be a sequence of valid characters, followed by a sequence of invalid characters, followed by a sequence of characters with the 8th bit set
    for check_hash in t.keys():
        if check_hash in t:
            array = t[check_hash]
            i = 0
            while i < len(array) and valid.find(chr(array[i])) == -1:
                i += 1
            while i < len(array) and valid.find(chr(array[i])) != -1:
                i += 1
            while i < len(array) and array[i] >= 128:
                i += 1
            if i < len(array):
                return False
        else:
            continue
    return True

#print_layout()
print(is_valid())
#sys.exit(0)
valid = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"
print("----")

def validate_top_bit():
    for src in range(256):
        for i in range(8):
            shifted = (src >> 3)%256
            shifted += i << 5
            check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
            if(check_hash & 0x80 != 0):
                if(src < 128 and valid.find(chr(src)) == -1):
                    #print("error ", hex(src), hex(check_hash))
                    return False
    return True
def fun_adjust():
    for zz in range(256):
        check_asso[ord('-')&0xf] = zz
        for yy in range(256):
            check_asso[ord('_')&0xf] = yy
            if(is_valid() and validate_top_bit()):
                print("----")
                print_layout()
                print(is_valid())
                print("found")
                return
fun_adjust()
def adjust(array, start, end, check_hash):
    for j in range(256):
        is_ok = True
        for i in range(len(array)):
            valid = (sat(j,array[i])&0x80 == 0)
            should_be_valid = (i>=start and i < end)
            is_ok = is_ok and (valid == should_be_valid)
        if(is_ok):
            check_values[check_hash&0xf] = j
            return
    raise "unexpected"


    chk = sat(quietlookup(check_values,check_hash), src)
    mask = chk & 0x80

def process_explicit():
    t={}
    for i in range(256):
        src = i
        for j in range(8):
            shifted = (src >> 3)%256
            shifted += j << 5
            check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
            if check_hash not in t:
                t[check_hash] = []
            t[check_hash].append(i)
    for check_hash in t.keys():
        if check_hash in t:
            array = t[check_hash]
            i = 0
            while i < len(array) and valid.find(chr(array[i])) == -1:
                i += 1
            if i < len(array) and valid.find(chr(array[i])) != -1:
                start = i
                while i < len(array) and valid.find(chr(array[i])) != -1:
                    i += 1
                end = i
                adjust(array, start, end, check_hash)
        else:
            continue
    return True
print("process")
process_explicit()
print("string")
print(computestring()+ " "+str(len(computestring())))

for c in valid:
    print(c,processquiet(ord(c)))

def examine():
    t={}
    for i in valid:
        src = ord(i)
        shifted = (src >> 3)%256
        check_hash = (quietlookup(delta_asso,src) + shifted + 1) >> 1
        if check_hash not in t:
            t[check_hash] = []
        t[check_hash].append(i)
    for check_hash in t.keys():
        print(check_hash, t[check_hash])
    return True
examine()

delta_values[10] += 1 

delta_values[13] += 33 

for c in valid:
    print(c,processquiet(ord(c)))



def casthex(v):
    if(v >= 0x80):
        return "uint8_t("+"0x{:X}".format(v)+")"
    return "0x{:X}".format(v)
def printme(c):
    print(",".join([casthex(i) for i in c]))
print("delta_asso")
printme(delta_asso)
print("check_asso")
printme(check_asso)
print("delta_values")
printme(delta_values)
print("check_values")
printme(check_values)

def is_base64(src):
    is_base64 = None
    for i in range(8):
        shifted = (src >> 3)%256
        shifted += i << 5
        check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
        chk = sat(quietlookup(check_values,check_hash), src)
        this_is_base64 = chk & 0x80 == 0
        if is_base64 is None:
            is_base64 = this_is_base64
        else:
            if is_base64 != this_is_base64:
                return None
    return is_base64


def validate_top_bit():
    for src in range(256):
        for i in range(8):
            shifted = (src >> 3)%256
            shifted += i << 5
            check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
            if(check_hash & 0x80 != 0):
                if(src < 128 and valid.find(chr(src)) == -1):
                    #print("error ", hex(src), hex(check_hash))
                    return False
    return True

for i in range(256):
    if is_base64(i):
        print(chr(i), end=",")
    elif is_base64(i) is None:
        print("None", hex(i), end=",")



lookup_0 = [0 for i in range(64)]
lookup_1 = [0 for i in range(64)]
for i in range(64):
    lookup_0[i] = 0x80
    lookup_1[i] = 0x80
lookup = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
for i in range(64):
    val = ord(lookup[i])
    bit6 = val & 0x40
    bits05 = val & 0x3f
    if bit6:
        lookup_1[bits05] = i
    else:
        lookup_0[bits05] = i
allowed = "\t\r\n "
for z in allowed:
    lookup_0[ord(z)] = 0xff
def sign8(x):
    if x >= 128:
        return x - 256
    return x
lookup_0.reverse()
lookup_1.reverse()
print("lookup0:")
print(", ".join([str(sign8(i)) for i in lookup_0]))
print("lookup1:")
print(", ".join([str(sign8(i)) for i in lookup_1]))
lookupn = [0 for i in range(64)]
output = 0
for ifrom in range(16):
    lookupn[ifrom*4 + 0] = output + 3
    lookupn[ifrom*4 + 1] = output + 2
    lookupn[ifrom*4 + 2] = output + 1
    lookupn[ifrom*4 + 3] = output + 0
    output += 4
lookupn.reverse()
print("reverse:")
print(", ".join([str(i) for i in lookupn]))

print("====")

lookup_0 = [0 for i in range(64)]
lookup_1 = [0 for i in range(64)]
for i in range(64):
    lookup_0[i] = 0x80
    lookup_1[i] = 0x80
lookup = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"
for i in range(64):
    val = ord(lookup[i])
    bit6 = val & 0x40
    bits05 = val & 0x3f
    if bit6:
        lookup_1[bits05] = i
    else:
        lookup_0[bits05] = i
allowed = "\0\t\r\n "
for z in allowed:
    lookup_0[ord(z)] = 0xff

lookup_0.reverse()
lookup_1.reverse()
print("lookup0:")
print(", ".join([str(sign8(i)) for i in lookup_0]))
print("lookup1:")
print(", ".join([str(sign8(i)) for i in lookup_1]))
lookupn = [0 for i in range(64)]
output = 0
for ifrom in range(16):
    lookupn[ifrom*4 + 0] = output + 3
    lookupn[ifrom*4 + 1] = output + 2
    lookupn[ifrom*4 + 2] = output + 1
    lookupn[ifrom*4 + 3] = output + 0
    output += 4
lookupn.reverse()
print("reverse:")
print(", ".join([str(i) for i in lookupn]))


#!/usr/bin/env python3

from pathlib import Path
import sys
import argparse
import json
import subprocess
import re


def main():
    args = parse_args()

    benchmark = args.build_dir / 'benchmarks' / 'benchmark'
    if not benchmark.is_file():
        raise RuntimeError(f"{benchmark} not found, was simdutf compiled with `-DSIMDUTF_BENCHMARKS=ON`?")

    # load registered tests
    simdutf_tests = [entry for entry in json.loads(execute([benchmark, "--json"])) if entry["simdutf"]]
    maxlen = max(len(entry['name']) for entry in simdutf_tests)
    count = len(simdutf_tests)

    for id, test in enumerate(simdutf_tests, 1):
        name = test["name"]
        encodings = test["encodings"]

        print(f"[{id}/{count}] {name:{maxlen}}... ", end='')
        sys.stdout.flush()

        if args.filter and not re.search(args.filter, name):
            print(f"skipping, does not match filter '{args.filter}'")
            continue

        if len(encodings) > 1:
            print(f"skipping, requires multiple input encodings ({encodings})")
            continue

        report_name = (name + '.txt')
        report_path = args.output_dir / report_name
        if report_path.exists():
            print("already benchmarked")
            continue

        enc = encodings[0]
        cmd = [benchmark, "-P", name, "-F"] + input_files(args.lipsum_dir, enc)
        result = execute(cmd)
        print(f"creating {report_path}")
        report_path.write_bytes(result)


def input_files(lipsum_dir, encoding):
    dir = lipsum_dir / 'wikipedia_mars'
    return list(dir.glob(patterns[encoding]))


# in unicode_ipsum/wikipedia_mars
patterns = {
    'latin1'    : '*.latin1.txt',
    'utf8'      : '*.utf8.txt',
    'utf16le'   : '*.utf16.txt',
    'utf16be'   : '*.utf16be.txt',
    'utf32le'   : '*.utf32.txt',
}


def parse_args():
    p = argparse.ArgumentParser("SIMDUTF tool for benchmarking")
    p.add_argument("--build-dir",
                   required=True,
                   type=existing_dir,
                   metavar="DIR",
                   help="simdutf build directory")
    p.add_argument("--lipsum-dir",
                   required=True,
                   type=existing_dir,
                   metavar="DIR",
                   help="unicode lorem lipsum directory")
    p.add_argument("--output-dir",
                   required=True,
                   type=existing_dir,
                   metavar="DIR",
                   help="output directory")
    p.add_argument("--filter",
                   type=str,
                   metavar="PATTERN",
                   help="only run benchmarks whose name matches this regex pattern")

    args = p.parse_args()

    return args


def existing_dir(s):
    path = Path(s).expanduser().absolute()
    if not path.is_dir():
        raise argparse.ArgumentError("'{path}' is not a directory")

    return path


def execute(params):
    cmd = [str(param) for param in params]
    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if result.returncode != 0:
        print(' '.join(cmd))
        if result.stderr:
            print("stderr:")
            print(result.stderr)
        if result.stdout:
            print("stdout:")
            print(result.stdout)
        print(f"return code: {result.returncode}")
        raise ValueError("command failed")

    return result.stdout


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Script is used to produce file `ppc64_utf16_to_utf8_tables.h`, that
contains lookup tables used in UTF-16 -> UTF-8 translation. Lookup
tables provide shuffle pattern for packing words into 1, 2, or 3-byte
UTF-8 characters.
"""

import sys
from os.path import normpath
from pathlib import Path
from contextlib import redirect_stdout


def format_array(array):
    result = []
    for value in array:
        if value < 0 or value == 0x80:
            result.append('0x80')
        else:
            result.append(str(value))

    return ', '.join(result)


def assure_array_length(array, size, value=0x80):
    while len(array) < size:
        array.append(value)


LE = True
BE = False


def shuffle_for_conversion_1_2_3_utf8_bytes(endianness):
    # There are two 8-bit bitmask telling how many bytes each word produces (1, 2 or 3).
    # mask1 = ddccbbaa -- output exactly one byte (d - MSB, a - LSB)
    # mask2 = hhggffee -- output one or two bytes

    # Please note that each bit is duplicated. In final form these bits are interleaved:
    # mask  = (mask1 & 0x5555) | (mask2 & 0xaaaa)
    #       = hdgcfbea

    # Each two-bit subword decides how many bytes will be copied from a 32-bit word of register:
    # | e | a | ea |
    # +---+---+----+-------
    # | 0 | 0 |  0 |  3 bytes
    # | 0 | 1 |  1 |  -- such combination will never come from C++ code, it has no sense
    # | 1 | 0 |  2 |  2 bytes
    # | 1 | 1 |  3 |  1 byte

    # unlike SSE, the PPC64 shuffle reads from **two** 16-byte vectors
    if endianness == BE:
        byte0_3bytes = 1
        byte1_3bytes = 0
        byte2_3bytes = 16 + 0

        byte0_2bytes = 0
        byte1_2bytes = 16 + 0

        byte0_1byte = 16 + 1
    else:
        byte0_3bytes = 0
        byte1_3bytes = 1
        byte2_3bytes = 16 + 1

        byte0_2bytes = 1
        byte1_2bytes = 16 + 1

        byte0_1byte = 16 + 0

    for mask in range(256):
        empty = 0x80
        shuffle = []
        for i in range(4):
            subword = mask & 0b11
            mask >>= 2

            if subword == 0:
                shuffle.append(i*2 + byte0_3bytes)
                shuffle.append(i*2 + byte1_3bytes)
                shuffle.append(i*2 + byte2_3bytes)
            elif subword == 3:
                shuffle.append(i*2 + byte0_1byte)
            elif subword == 2:
                shuffle.append(i*2 + byte0_2bytes)
                shuffle.append(i*2 + byte1_2bytes)

        output_bytes = len(shuffle)
        assure_array_length(shuffle, 16, empty)

        yield (shuffle, output_bytes)


CPP_HEADER = """// Code generated automatically; DO NOT EDIT
// file generated by scripts/ppc64_convert_utf16_to_utf8.py
#ifndef PPC64_SIMDUTF_UTF16_TO_UTF8_TABLES_H
#define PPC64_SIMDUTF_UTF16_TO_UTF8_TABLES_H

namespace simdutf {
namespace {
namespace tables {
namespace ppc64_utf16_to_utf8 {
"""

CPP_FOOTER = """} // ppc64_utf16_to_utf8 namespace
} // tables namespace
} // unnamed namespace
} // namespace simdutf

#endif // PPC64_SIMDUTF_UTF16_TO_UTF8_TABLES_H"""


def lookup_3_byte_utf8_for_perm():
    byte0_3bytes = 1
    byte1_3bytes = 0
    byte2_3bytes = 16 + 0

    tmp = []
    for i in range(8):
        tmp.append(2*i + byte0_3bytes)
        tmp.append(2*i + byte1_3bytes)
        tmp.append(2*i + byte2_3bytes)

    print("lo =", tmp[:16])
    print("hi =", tmp[16:])


def main():
    name = 'ppc64_utf16_to_utf8_tables.h'
    path = normpath(Path(__file__).parent / '..' / 'src' / 'ppc64' / name)
    print(f"creating {path}")
    with open(path, 'wt') as f:
        with redirect_stdout(f):
            print(CPP_HEADER)

            print("#if SIMDUTF_IS_BIG_ENDIAN")
            print("// 1 byte for length, 16 bytes for mask")
            print("const uint8_t pack_1_2_3_utf8_bytes[256][17] = {")
            for shuffle, size in shuffle_for_conversion_1_2_3_utf8_bytes(BE):
                print("  {%s}," % format_array([size] + shuffle))
            print("};")
            print("#else")
            print("// 1 byte for length, 16 bytes for mask")
            print("const uint8_t pack_1_2_3_utf8_bytes[256][17] = {")
            for shuffle, size in shuffle_for_conversion_1_2_3_utf8_bytes(LE):
                print("  {%s}," % format_array([size] + shuffle))
            print("};")
            print("#endif // SIMDUTF_IS_BIG_ENDIAN")

            print(CPP_FOOTER)


if __name__ == '__main__':
    main()



import sys
delta_asso = [0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00,0x00, 0x00, 0x00, 0x0F, 0x00, 0x0F]
check_asso = [0x0D, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x03, 0x07, 0x0B, 0x0B, 0x0B, 0x0F]

delta_values =[(0x00), (0x00), (0x00), (0x13), (0x04), (0xBF), (0xBF), (0xB9), (0xB9), (0x00), (0x10), (0xC3), (0xBF), (0xBF), (0xB9), (0xB9)]
check_values = [(0x80), (0x80), (0x80), (0x80), (0xCF), (0xBF), (0xD5), (0xA6), (0xB5), (0x86), (0xD1), (0x80), (0xB1), (0x80), (0x91), (0x80)]


valid = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"

def safechr(i):
    if i < 32:
        return '.'
    if i > 127:
        return '?'
    return chr(i)

def safehex(x):
    return "0x{0:2x}".format(x)

def to_signed(x):
    if(x >= 128):
        return x - 256
    return x

def to_unsigned(x):
    if(x < 0):
        return x + 256
    return x

def sat(x, y):
    x = to_signed(x)
    y = to_signed(y)
    z = x + y
    if(z > 127):
        return 127
    if(z < -128):
        return to_unsigned(-128)
    return to_unsigned(z)

def lookup(table, index):
    print("looking up ", hex(index))
    if(index >= 128):
        return 0
    return table[index&0xf]



def quietlookup(table, index):
    if(index >= 128):
        return 0
    return table[index&0xf]

def process(src):
    shifted = (src >> 3)%256
    delta_hash = (lookup(delta_asso,src) + shifted + 1) >> 1
    check_hash = (lookup(check_asso,src) + shifted + 1) >> 1
    out = sat(lookup(delta_values,delta_hash), src)
    chk = sat(lookup(check_values,check_hash), src)
    mask = chk & 0x80
    return (out, mask)

def processquiet(src):
    shifted = (src >> 3)%256
    delta_hash = (quietlookup(delta_asso,src) + shifted + 1) >> 1
    check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
    out = sat(quietlookup(delta_values,delta_hash), src)
    chk = sat(quietlookup(check_values,check_hash), src)
    mask = chk & 0x80
    return (out, mask)

def is_ok(i):
    out, mask = processquiet(i)
    if mask == 0:
        return 1
    return 0

def computestring():
    s = ""
    for i in range(256):
        out, mask = processquiet(i)
        if(mask == 0):
            s +=  safechr(i)
    return s
print(computestring() + " " + str(len(computestring())))

def print_layout():
    t={}
    for i in range(256):
        src = i
        shifted = (src >> 3)%256
        check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
        if check_hash not in t:
            t[check_hash] = []
        t[check_hash].append(i)
    for check_hash in range(16):
        if check_hash in t:
            off = quietlookup(check_values,check_hash)
            print(hex(check_hash), hex(off), end="")
            print("\t", " ".join(["   "+safechr(c) for c in t[check_hash]]))
        else:
            continue


def is_valid():
    t={}
    for i in range(256):
        src = i
        shifted = (src >> 3)%256
        check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
        if check_hash not in t:
            t[check_hash] = []
        t[check_hash].append(i)
    for check_hash in t.keys():
        if check_hash in t:
            array = t[check_hash]
            i = 0
            while i < len(array) and valid.find(chr(array[i])) == -1:
                i += 1
            while i < len(array) and valid.find(chr(array[i])) != -1:
                i += 1
            while i < len(array) and array[i] >= 128:
                i += 1
            if i < len(array):
                return False
        else:
            continue
    return True

print_layout()
print(is_valid())


def is_base64(src):
    is_base64 = None
    for i in range(8):
        shifted = (src >> 3)%256
        shifted += i << 5
        #print("shifted ", hex(shifted))
        delta_hash = (quietlookup(delta_asso,src) + shifted + 1) >> 1
        #print("delta_hash ", hex(delta_hash))
        check_hash = (quietlookup(check_asso,src) + shifted + 1) >> 1
        #print("check_hash ", hex(check_hash))
        out = sat(quietlookup(delta_values,delta_hash), src)
        #print("out ", hex(out))
        chk = sat(quietlookup(check_values,check_hash), src)
        this_is_base64 = chk & 0x80 == 0
        if is_base64 is None:
            is_base64 = this_is_base64
        else:
            if is_base64 != this_is_base64:
                print("unexpected")
                return None
    return is_base64

for i in range(256):
    if is_base64(i):
        print(chr(i), end=",")
    elif is_base64(i) is None:
        print("None", hex(i), end=",")


t='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'
spaces=' \t\n\r\f'
lut_lo = [0x3a, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x61, 0xe1, 0xb4, 0xe5, 0xe5, 0xf4, 0xb4]
lut_hi = [0x11, 0x20, 0x42, 0x80, 0x8,  0x4,  0x8,  0x4, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20]
roll = [0x0, 0x10, 0x13, 0x4, 0xbf, 0xbf, 0xb9, 0xb9, 0x0, 0x0,  0x0,  0x0, 0x0,  0x0,  0x0,  0x0]
def decode(s):
    low = s & 0xf
    high = s >> 4
    m = lut_lo[low] & lut_hi[high]
    if(m > 0x3):
        return (m, None)
    if s == 0x2f:
        off = roll[high - 1]
    else:
        off = roll[high]
    return (m,(s + off)&0xff)


for i in range(256):
    m,d = decode(i)
    if d is None:
        assert t.find(chr(i)) == -1
        assert spaces.find(chr(i)) == -1
        continue
    if m == 0:
        assert d >= 0
        # we must have a base64 element
        v = t.find(chr(i))
        assert v == d
    else:
        # we must have a space
        v = spaces.find(chr(i))
        assert v >= 0, (i, chr(i), v)
print("REGULAR")
print("lut_lo")
print(",".join([hex(c) for c in lut_lo]))
print("lut_hi")
print(",".join([hex(c) for c in lut_hi]))
print("roll")
print(",".join([hex(c) for c in roll]))


print("SWITCHING TO BASE64URL")
print()
t='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'
spaces=' \t\n\r\f'

lut_lo = [0x3a, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x70, 0x61, 0xe1, 0xb4, 0xf4, 0xe5, 0xf4, 0xb0]
lut_hi = [0x11, 0x20, 0x42, 0x80, 0x8,  0x4,  0x8,  0x4, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20]
roll = [0xe0, 0x11, 0x13, 0x4, 0xbf, 0xbf, 0xb9, 0xb9, 0x0, 0x0,  0x0,  0x0, 0x0,  0x0,  0x0,  0x0]
t='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'
spaces=' \t\n\r\f' ## ['0x20', '0x9', '0xa', '0xd']


lut_lo = [0x0 for i in range(16)]
lut_hi = [0x0 for i in range(16)]
#roll = [0 for i in range(16)]

#0x00 are forbidden except for \t \n \r \f which go to one
lut_hi[0] = 0x11
for z in range(16):
    if '\t\n\r\f'.find(chr(z)) != -1:
        lut_lo[z & 0xf] = 0x1 # allowed
    else:
        lut_lo[z] = 0x10 # forbidden
#0x10 and 0x80 all forbidden
lut_hi[0x1] = 0x20
for z in range(0x8, 16):
    lut_hi[z] = 0x20
#lut_hi[0x8] = 0x20

for z in range(16):
    lut_lo[z] |= 0x20

#0x20 selective
lut_hi[0x2] = 0x42
for z in range(16):
    if z == 0:
        lut_lo[z] |= 0x2
    elif z != 0xd:
        lut_lo[z] |= 0x40


#0x30 numbers
lut_hi[0x3] = 0x80
for z in range(10,16):
    lut_lo[z] |= 0x80

#0x40, 0x60 letters
lut_hi[0x4] = 0x8
lut_hi[0x6] = 0x8
lut_lo[0] |= 0x8

#0x7 letters
#0x5 letters
lut_hi[0x5] |= 0x4
lut_hi[0x7] |= 0x4
for i in range(0xb,16):
    lut_lo[i] |= 0x4



def decodes(s):
    low = s & 0xf
    high = s >> 4
    m = lut_lo[low] & lut_hi[high]
    is_underscore = s == 0x5f
    if(is_underscore):
        m = 0
        high = 0
    if(m > 0x3):
        return (m, None)
    if s == 0x2d:
        off = roll[high - 1]
    else:
        off = roll[high]
    return (m,(s + off)&0xff)
print("lut_lo")
print(",".join([hex(c) for c in lut_lo]))
print("lut_hi")
print(",".join([hex(c) for c in lut_hi]))
print("roll")
print(",".join([hex(c) for c in roll]))


for i in range(256):
    m,d = decodes(i)
    if d is None:
        assert t.find(chr(i)) == -1
        assert spaces.find(chr(i)) == -1
        continue
    if m == 0:
        assert d >= 0
        # we must have a base64 element
        v = t.find(chr(i))
        if(v != d): 
            print(hex(i), chr(i), v, d)
        #assert v == d
    else:
        # we must have a space
        v = spaces.find(chr(i))
        assert v >= 0, (i, chr(i), v)




class TableBase(object):
    def __init__(self):
        self.headers = []
        self.rows    = []

    def set_header(self, header):
        assert len(header) > 0
        self.headers = [self.normalize(header)]

    def add_row(self, row):
        assert len(row) > 0
        self.rows.append(self.normalize(row))

    def normalize(self, row):
        tmp = []
        for text, count in self.iter_spans(row):
            assert count >= 1
            if count == 1:
                tmp.append(text)
            else:
                tmp.append((text, count))

        return tmp

    def is_raw(self, row):
        return all(isinstance(val, str) for val in row)

    def iter_spans(self, row):
        for item in row:
            if type(item) is tuple:
                text  = item[0]
                count = item[1]
            else:
                text  = item
                count = 1

            yield (text, count)


class TableValidator(object):
    def __init__(self, table):
        self.table = table
        self.columns = self.get_table_columns_count()
        if self.columns is None:
            raise ValueError("Table doesn't have header which define column layout")
        self.validate()

    def get_table_columns_count(self):
        # only from headers
        for header in self.table.headers:
            if self.table.is_raw(header):
                return len(header)

    def validate(self):
        for i, header in enumerate(self.table.headers):
            n = self.get_columns_count(header)
            if n != self.columns:
                raise ValueError("header #%d has %d column(s), expected %d: %s" % (i, n, self.columns, header))

        for i, row in enumerate(self.table.rows):
            n = self.get_columns_count(row)
            if n != self.columns:
                raise ValueError("row #%d has %d column(s), expected %d: %s" % (i, n, self.columns, row))

    def get_columns_count(self, row):
        n = 0
        for _, count in self.table.iter_spans(row):
            n += count

        return n


ALIGN_RIGHT = '>'
ALIGN_LEFT = '<'
CENTER = '^'


class MarkdownTableRenderer(object):
    def __init__(self, table):
        if len(table.headers) > 1:
            raise ValueError("Markdown does not support multiple header rows")

        self.validator  = TableValidator(table)
        self.table      = table
        self.padding    = 1
        self.widths     = self._calculate_widths()

    def get_headers(self):
        return self.table.headers

    def get_rows(self):
        return self.table.rows

    def _calculate_widths(self):
        width = [0] * self.validator.columns

        # get width from fixed
        for row in self.get_headers() + self.get_rows():
            index = 0
            for text, count in self.table.iter_spans(row):
                w = len(text)
                width[index] = max(w, width[index])
                index += 1

        return width

    def _get_columns_width(self, start, count):
        assert count >= 1

        w = 0
        for index in range(start, start + count):
            w += self.widths[index]
            w += 2 * self.padding

        w += (count - 1)  # for the columns spacing '|'
        return w

    def _render_separator(self, row, fill):
        assert len(fill) == 1

        result = '|'
        index = 0
        for text, count in self.table.iter_spans(row):
            result += ' '
            result += fill * (self._get_columns_width(index, count) - 2)
            result += ' '
            result += '|'
            index  += count

        return result

    def _render_row(self, row, align=None):
        if align is not None:
            def get_align(text):
                return align
        else:
            def get_align(text):
                if is_float_or_int(text):
                    return ALIGN_RIGHT
                else:
                    return ALIGN_LEFT

        result = '|'
        index = 0
        for text, count in self.table.iter_spans(row):
            width   = self._get_columns_width(index, count)
            width  -= 2 * self.padding
            result += ' ' * self.padding
            result += u'{:{align}{width}}'.format(text, align=get_align(text), width=width)
            result += ' ' * self.padding
            result += '|'

            index  += count

        return result

    def get_image(self):
        lines = []

        if self.table.headers:
            header = self.table.headers[0]
            lines.append(self._render_row(header, CENTER))
            lines.append(self._render_separator(header, '-'))

        for row in self.get_rows():
            lines.append(self._render_row(row))

        return '\n'.join(lines)


class Table(TableBase):
    def __unicode__(self):
        renderer = MarkdownTableRenderer(self)
        return renderer.get_image()

    def __str__(self):
        renderer = MarkdownTableRenderer(self)
        return renderer.get_image()


def is_float_or_int(text):
    try:
        float(text)
        return True
    except ValueError:
        return False


if __name__ == '__main__':

    table = Table()

    table.set_header(["procedure", "size", "time"])
    table.add_row(["foo", "100", "0.5"])
    table.add_row(["bar", "105", "1.5"])
    table.add_row(["baz", "111", "0.2"])

    print(table)

    table2 = Table()
    table2.set_header(["size1", "size2", "proc1", "proc2", "proc3"])
    table2.add_row(["1", "2", "a", "b", "c"])
    table2.add_row(["9", "3", "A", "B", "C"])
    table2.add_row(["42", "-", ("N/A", 3)])

    print()
    print(table2)



#!/usr/bin/env python3
########################################################################
# Generates a new release.
########################################################################
import sys
import re
import subprocess
import io
import os
import fileinput
if sys.version_info < (3, 0):
    sys.stdout.write("Sorry, requires Python 3.x or better\n")
    sys.exit(1)

def colored(r, g, b, text):
    return "\033[38;2;{};{};{}m{} \033[38;2;255;255;255m".format(r, g, b, text)

def extractnumbers(s):
    return tuple(map(int,re.findall(r"(\d+)\.(\d+)\.(\d+)",str(s))[0]))

def toversionstring(major, minor, rev):
    return str(major)+"."+str(minor)+"."+str(rev)

def topaddedversionstring(major, minor, rev):
    return str(major)+str(minor).zfill(3)+str(rev).zfill(3)
print("Calling git rev-parse --abbrev-ref HEAD")
pipe = subprocess.Popen(["git", "rev-parse", "--abbrev-ref", "HEAD"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
branchresult = pipe.communicate()[0].decode().strip()

if(branchresult != "master"):
    print(colored(255, 0, 0, "We recommend that you release on master, you are on '"+branchresult+"'"))

ret = subprocess.call(["git", "remote", "update"])

if(ret != 0):
    sys.exit(ret)
print("Calling git log HEAD.. --oneline")
pipe = subprocess.Popen(["git", "log", "HEAD..", "--oneline"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
uptodateresult = pipe.communicate()[0].decode().strip()

if(len(uptodateresult) != 0):
    print(uptodateresult)
    sys.exit(-1)

pipe = subprocess.Popen(["git", "rev-parse", "--show-toplevel"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
maindir = pipe.communicate()[0].decode().strip()
scriptlocation = os.path.dirname(os.path.abspath(__file__))

print("repository: "+maindir)

pipe = subprocess.Popen(["git", "describe", "--abbrev=0", "--tags"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
versionresult = pipe.communicate()[0].decode().strip()

print("last version: "+versionresult )
try:
  currentv = extractnumbers(versionresult)
except:
  currentv = [0,0,0]
if(len(sys.argv) != 2):
    nextv = (currentv[0],currentv[1], currentv[2]+1)
    print ("please specify version number, e.g. "+toversionstring(*nextv))
    sys.exit(-1)
try:
    newversion = extractnumbers(sys.argv[1])
    print(newversion)
except:
    print("can't parse version number "+sys.argv[1])
    sys.exit(-1)

print("checking that new version is valid")
if(newversion[0] !=  currentv[0]):
    assert newversion[0] >=  currentv[0]
    assert newversion[1] == 0
    assert newversion[2] == 0
elif (newversion[1] !=  currentv[1]):
    assert newversion[1] >=  currentv[1]
    assert newversion[2] == 0
else :
    assert newversion[2] >=  currentv[2]

atleastminor= (currentv[0] != newversion[0]) or (currentv[1] != newversion[1])


versionfilerel = os.sep + "include" + os.sep + "simdutf" + os.sep + "simdutf_version.h"
versionfile = maindir + versionfilerel

with open(versionfile, 'w') as file:
    file.write("// /include/simdutf/simdutf_version.h automatically generated by release.py,\n")
    file.write("// do not change by hand\n")
    file.write("#ifndef SIMDUTF_SIMDUTF_VERSION_H\n")
    file.write("#define SIMDUTF_SIMDUTF_VERSION_H\n")
    file.write("\n")
    file.write("/** The version of simdutf being used (major.minor.revision) */\n")
    file.write("#define SIMDUTF_VERSION \""+toversionstring(*newversion)+"\"\n")
    file.write("\n")
    file.write("namespace simdutf {\n")
    file.write("enum {\n")
    file.write("  /**\n")
    file.write("   * The major version (MAJOR.minor.revision) of simdutf being used.\n")
    file.write("   */\n")
    file.write("  SIMDUTF_VERSION_MAJOR = "+str(newversion[0])+",\n")
    file.write("  /**\n")
    file.write("   * The minor version (major.MINOR.revision) of simdutf being used.\n")
    file.write("   */\n")
    file.write("  SIMDUTF_VERSION_MINOR = "+str(newversion[1])+",\n")
    file.write("  /**\n")
    file.write("   * The revision (major.minor.REVISION) of simdutf being used.\n")
    file.write("   */\n")
    file.write("  SIMDUTF_VERSION_REVISION = "+str(newversion[2])+"\n")
    file.write("};\n")
    file.write("} // namespace simdutf\n")
    file.write("\n")
    file.write("#endif // SIMDUTF_SIMDUTF_VERSION_H\n")

print(versionfile + " modified")

newmajorversionstring = str(newversion[0])
mewminorversionstring = str(newversion[1])
newrevversionstring = str(newversion[2])
newversionstring = str(newversion[0]) + "." + str(newversion[1]) + "." + str(newversion[2])
cmakefile = maindir + os.sep + "CMakeLists.txt"
sonumber = None
pattern = re.compile(r"""set\(SIMDUTF_LIB_SOVERSION "(\d+)" CACHE STRING "simdutf library soversion"\)""")
with open (cmakefile, 'rt') as myfile:
    for line in myfile:
        m = pattern.search(line)
        if m is not None:
            sonumber = int(m.group(1))
            break
print("so library number "+str(sonumber))

if(atleastminor):
    print("Given that we have a minor revision, it seems necessary to bump the so library number")
    if sonumber is None:
        print("I cannot find the so library number in the CMakeLists.txt file")
        sys.exit(-1)
    sonumber += 1

for line in fileinput.input(cmakefile, inplace=True, backup='.bak'):
    line = re.sub(r'  VERSION \d+\.\d+\.\d+','  VERSION '+newmajorversionstring+'.'+mewminorversionstring+'.'+newrevversionstring, line.rstrip())
    line = re.sub(r'SIMDUTF_LIB_VERSION "\d+\.\d+\.\d+','SIMDUTF_LIB_VERSION "'+str(sonumber)+".0.0", line)
    line = re.sub(r'set\(SIMDUTF_LIB_SOVERSION "\d+"','set(SIMDUTF_LIB_SOVERSION \"'+str(sonumber)+'\"', line)
    print(line)

print("modified "+cmakefile+", a backup was made")


doxyfile = maindir + os.sep + "Doxyfile"
for line in fileinput.input(doxyfile, inplace=True, backup='.bak'):
    line = re.sub(r'PROJECT_NUMBER         = "\d+\.\d+\.\d+','PROJECT_NUMBER         = "'+newversionstring, line.rstrip())
    print(line)
print("modified "+doxyfile+", a backup was made")


print("running amalgamate.py")
cp = subprocess.run(["python3", maindir+ os.sep + "singleheader/amalgamate.py"], stdout=subprocess.DEVNULL)  # doesn't capture output

if(cp.returncode != 0):
    print("Failed to run amalgamate")

print("running doxygen")
cp = subprocess.run(["doxygen"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=maindir)  # doesn't capture output

if(cp.returncode != 0):
    print("Failed to run doxygen")


readmefile = maindir + os.sep + "README.md"


for line in fileinput.input(readmefile, inplace=True, backup='.bak'):
    line = re.sub(r'   wget https://github.com/simdutf/simdutf/releases/download/v\d+\.\d+\.\d+/singleheader.zip','   wget https://github.com/simdutf/simdutf/releases/download/v'+newmajorversionstring+'.'+mewminorversionstring+'.'+newrevversionstring+'/singleheader.zip', line.rstrip())
    line = re.sub(r'https://github.com/simdutf/simdutf/releases/download/v\d+\.\d+\.\d+/singleheader.zip','https://github.com/simdutf/simdutf/releases/download/v'+newmajorversionstring+'.'+mewminorversionstring+'.'+newrevversionstring+'/singleheader.zip', line.rstrip())
    print(line)

print("modified "+readmefile+", a backup was made")

pattern = re.compile(r"https://simdutf.org/api/(\d+\.\d+\.\d+)/index.html")
readmedata = open(readmefile).read()
m = pattern.search(readmedata)
if m == None:
    print('I cannot find a link to the API documentation in your README')
else:
    if m is not None:
        detectedreadme = m.group(1)
        print("found a link to your API documentation in the README file: "+detectedreadme+" ("+toversionstring(*newversion)+")")
        if(atleastminor):
            if(detectedreadme != toversionstring(*newversion)):
                print(colored(255, 0, 0, "Consider updating the readme link to "+toversionstring(*newversion)))



print("Please run the tests before issuing a release. \n")
print("to issue release, enter \n git commit -a && git push  &&  git tag -a v"+toversionstring(*newversion)+" -m \"version "+toversionstring(*newversion)+"\" &&  git push --tags \n")






#!/usr/bin/env python3

from contextlib import redirect_stdout


def format_array(array):
    result = []
    for value in array:
        if value < 0 or value == 0x80:
            result.append('0x80')
        else:
            result.append(str(value))

    return ', '.join(result)


def assure_array_length(array, size, value=0x80):
    while len(array) < size:
        array.append(value)


def utf32_to_utf16le():
    for mask in range(16):
        src = 0
        arr = []
        for bit in [0x01, 0x02, 0x04, 0x08]:
            if mask & bit:
                arr.append(src + 0)
                arr.append(src + 1)
                arr.append(src + 2)
                arr.append(src + 3)
            else:
                arr.append(src + 0)
                arr.append(src + 1)

            src += 4

        assure_array_length(arr, 16)

        yield arr


def utf32_to_utf16be():
    for mask in range(16):
        src = 0
        arr = []
        for bit in [0x01, 0x02, 0x04, 0x08]:
            if mask & bit:
                arr.append(src + 1)
                arr.append(src + 0)
                arr.append(src + 3)
                arr.append(src + 2)
            else:
                arr.append(src + 1)
                arr.append(src + 0)

            src += 4

        assure_array_length(arr, 16)

        yield arr


CPP_HEADER = """// file generated by scripts/sse_convert_utf32_to_utf16.py
#ifndef SIMDUTF_UTF32_TO_UTF16_TABLES_H
#define SIMDUTF_UTF32_TO_UTF16_TABLES_H

namespace simdutf {
namespace {
namespace tables {
namespace utf32_to_utf16 {
"""

CPP_FOOTER = """} // utf16_to_utf8 namespace
} // tables namespace
} // unnamed namespace
} // namespace simdutf

#endif // SIMDUTF_UTF16_TO_UTF8_TABLES_H
"""


def main():
    with open('utf32_to_utf16_tables.h', 'wt') as f:
        with redirect_stdout(f):
            print(CPP_HEADER)

            print("const uint8_t pack_utf32_to_utf16le[16][16] = {")
            for row in utf32_to_utf16le():
                print("{%s}," % format_array(row))
            print("};")

            print()

            print("const uint8_t pack_utf32_to_utf16be[16][16] = {")
            for row in utf32_to_utf16be():
                print("{%s}," % format_array(row))
            print("};")

            print()
            print(CPP_FOOTER)


if __name__ == '__main__':
    main()



import textwrap
import sys
if sys.version_info[0] < 3:
    print('You need to run this with Python 3')
    sys.exit(1)

indent = ' ' * 4

def fill(text):
    tmp = textwrap.fill(text)
    return textwrap.indent(tmp, indent)

def filltab(text):
    tmp = textwrap.fill(text, width=120)
    return textwrap.indent(tmp, '\t')

def cpp_array_initializer(arr):
    return '{%s}' % (', '.join(map(str, arr)))

def compose2(f, g):
    return lambda x: f(g(x))

def cpp_arrayarray_initializer(arr):
    return '{%s}' % (',\n '.join(map(compose2(filltab,cpp_array_initializer), arr)))



# Note: Validation is done for 8-word input, we just need to check 3^8 = 6561 cases
#       Validation for 16-word inputs requires 3**16 = 43'046'721 checks

ELEMENTS_COUNT   = 8
ALL_MASK         = (1 << ELEMENTS_COUNT) - 1
ALL_BUT_ONE_MASK = (ALL_MASK >> 1)

# 'V' - single-word character (always valid)
# 'L' - low surrogate (must be followed by the high surrogate)
# 'H' - high surrogate
def all_sequences():
    index = ['V'] * ELEMENTS_COUNT

    def increment():
        nonlocal index
        for i in range(ELEMENTS_COUNT):
            if index[i] == 'V':
                index[i] = 'L'
                return False

            if index[i] == 'L':
                index[i] = 'H'
                return False

            if index[i] == 'H':
                index[i] = 'V' # wrap around
                pass

        return True

    overflow = False
    while not overflow:
        yield index
        overflow = increment()


def find_error_in_words(words):
    prev = None

    if words[0] == 'H':
        # We assume that our vector algorithm loads proper data into vectors.
        # In the case low surrogate was the last item in the previous iteration.
        return 'high surrogate must not start a chunk'

    for i, kind in enumerate(words):
        if kind == 'V':
            if prev == 'L':
                return f'low surrogate {i - 1} must be followed by high surrogate'
        elif kind == 'L':
            if prev == 'L':
                return f'low surrogate {i - 1} must be followed by high surrogate'
        elif kind == 'H':
            if prev != 'L':
                return f'high surrogate {i} must be preceded by low surrogate'

        prev = kind

    return ''


def bitmask(words, state):
    result = 0
    for bit, type in enumerate(words):
        if type == state:
            # In SSE vector algorithm we compare 2 x 16 higher bytes of input
            # words, which yields a 16-bit mask.
            result |= 1 << bit

    return result


def mask(words):
    L = bitmask(words, 'L')
    H = bitmask(words, 'H')
    V = (~(L | H)) & ALL_MASK

    a = L & (H >> 1)
    b = a << 1
    c = V | a | b

    return c


def dump():
    for words in all_sequences():

        c = mask(words)

        words_image = "[ %s ]" % ' | '.join(words)
        error = find_error_in_words(words)
        if error == '':
            valid_image = 'T'
        else:
            valid_image = ' '

        print(words_image, valid_image, '{:016b} {:04x}'.format(c, c))


def proof():
    case1_hit = False
    case2_hit = False
    for words in all_sequences():

        c = mask(words)

        if c == ALL_MASK:
            case1_hit = True
            # all 16 words are valid (either 'V' or pairs 'L', 'H')
            assert find_error_in_words(words) == '', (words, find_error_in_words(words))

        if c == ALL_BUT_ONE_MASK:
            case2_hit = True
            # all 15 words are valid (either 'V' or pairs 'L', 'H')
            # the last words is either 'L' or 'H' (the word will be
            # re-examined in the next iteration of an algorithm)
            if words[-1] == 'H':
                assert find_error_in_words(words) == 'high surrogate 7 must be preceded by low surrogate'
            elif words[-1] == 'L':
                assert find_error_in_words(words) == ''
            else:
                assert False

    assert case1_hit
    assert case2_hit

    print("All OK")


def main():
    if 0:
        dump()
    else:
        proof()


if __name__ == '__main__':
    main()



from itertools import product
from random import randint, seed
from sse_validate_utf16le_proof import bitmask


# This is a copy from sse_validate_utf16le_proof.py with
# adjusted the mask for the 16-bit base
def mask(words):
    L = bitmask(words, 'L')
    H = bitmask(words, 'H')
    V = (~(L | H)) & 0xffff

    a = L & (H >> 1)
    b = a << 1
    c = V | a | b

    return c


class Record:
    def __init__(self):
        self.words = []


    def add(self, word):
        self.words.append(word)


    @property
    def is_valid(self):
        c = mask(self.words)

        if c == 0xffff:
          return True

        if c == 0x7fff:
          # in test we reject cases when 'L' or 'H' ends a chunk
          if self.words[-1] in ('L', 'H'):
            return False
          else:
            return True

        return False


    def __str__(self):
        words = ''.join(self.words)
        if self.is_valid:
            return 'T' + words
        else:
            return 'F' + words


def test_words():
    collection = set()
    for seq in test_words_aux():
        collection.add(tuple(seq))

    return sorted(collection)


def test_words_aux():
    # 1. all valid
    yield ['V'] * 16

    # 2. only low surrogates
    yield ['L'] * 16

    # 3. only high surrogates
    yield ['H'] * 16

    # 4. sole low surrogate
    for i in range(16):
        seq = ['V'] * 16
        seq[i] = 'L'
        yield seq

    # 5. sole high surrogate
    for i in range(16):
        seq = ['V'] * 16
        seq[i] = 'H'
        yield seq

    # 6. scattered three surrogates
    for i in range(16):
        for j in range(16):
            for k in range(16):
                    seq = ['V'] * 16
                    for a, b, c in product('LH', repeat=3):
                        seq[i] = a
                        seq[j] = b
                        seq[k] = c
                        yield seq

    # To cover all 16-byte inputs we would need 3**16 cases (43'046'721)
    # Instead, we cover all possible 6-element combinations (3**6 = 729)
    # and move it within 16-element input. This yields 729 * 10 cases.
    k = 6
    for combination in product('VLH', repeat=k):
        for position in range(16 - k):
            seq = ['V'] * 16
            for i, v in enumerate(combination):
                seq[i + position] = v

            yield seq


TXT = """# generated by scripts/sse_validate_utf16le_testcases.py
"""

def write_file(file):
    file.write(TXT)
    for words in test_words():
        record = Record()
        for word in words:
            record.add(word)

        file.write(str(record))
        file.write('\n')


def main():
    seed(0)
    with open('validate_utf16_testcases.txt', 'w') as f:
        write_file(f)


if __name__ == '__main__':
    main()
